#! /usr/local/bin/expectk

lappend auto_path ../lib

package require twoOpt
package require twoIcons
package require twoMenu
package require twoToolbar
package require twoMdi
package require twoEditor

# ----------
# Global vars
# ----------

set Recent	""	;# List of recently opened files
set Clipboard	""	;# Application-internal clipboard
set posX	0	;# Cursor position: column
set posY	0	;# Cursor position: row

# ----------
# Menue Bar
# ----------

two_menu .m
. configure -menu .m

# ----------
# File menue
# ----------

# com = command; -com = -command; -x = -xaccel
two_menu .m.file -tearoff false -postcommand fileState
.m add cascade -menu .m.file -l "&File"                   -info "Open, save and print files, exit program"
.m.file add com -label "&New"        -com fileNew    -x n -info "Create a new file"
.m.file add com -label "&Open..."    -com fileOpen   -x o -info "Open an existing file"
.m.file add separator
.m.file add com -label "&Save"       -com fileSave   -x s -info "Save current file"
.m.file add com -label "Save &as..." -com fileSaveAs      -info "Save current file with a new name"
.m.file add com -label "&Close"      -com fileClose  -x w -info "Close current file"
.m.file add separator
.m.file add com -label "&Print"      -com filePrint  -x p -info "Print current file"
.m.file add separator
.m.file add com -label "E&xit"       -com fileExit   -x q -info "Exit program"

# set state of entries on menu open and append file history
proc fileState {} {
    global Recent

    set w [.mdi active].ed
    if { $w != ".ed" } {
        .m.file entryconfigure 3 -state [expr {[$w modState] ? "normal" : "disabled"}]
        .m.file entryconfigure 4 -state normal
        .m.file entryconfigure 5 -state normal
        .m.file entryconfigure 7 -state normal
    } else {
        .m.file entryconfigure 3 -state disabled
        .m.file entryconfigure 4 -state disabled
        .m.file entryconfigure 5 -state disabled
        .m.file entryconfigure 7 -state disabled
    }
    if {[.m.file index end] > 10} {
	.m.file delete 10 end
    }
    if {[llength $Recent] > 0} {
        .m.file add separator
	set idx 1
	foreach file $Recent {
	    .m.file add command -command "fileOpen $file" \
	        -label "&$idx: [file tail $file]" -info $file
	    incr idx
	}
    }
}

# additional binding for text
bind TwoText <Control-F4> fileClose

# ----------
# Edit menue
# ----------

# com = command; -com = -command; -x = -xaccel
two_menu .m.edit -tearoff false -postcommand editState
.m add cascade -menu .m.edit -label "&Edit"          -info "Edit current text"
.m.edit add com -label "&Undo"   -com editUndo  -x z -info "Undo last changes"
.m.edit add com -label "&Redo"   -com editRedo  -x r -info "Redo last undo"
.m.edit add separator
.m.edit add com -label "C&ut"    -com editCut   -x x -info "Cut selection and put it on the clipboard"
.m.edit add com -label "C&opy"   -com editCopy  -x c -info "Copy selection to the clipboard"
.m.edit add com -label "&Paste"  -com editPaste -x v -info "Paste contents of clipboard"

# set state of entries on menu open
proc editState {} {
    global Clipboard
    set w [.mdi active].ed
    if { $w != ".ed" } {
        .m.edit entryconfigure 0 -state [expr {[$w undoState]   ? "normal" : "disabled"}]
        .m.edit entryconfigure 1 -state [expr {[$w redoState]   ? "normal" : "disabled"}]
        .m.edit entryconfigure 3 -state [expr {[$w selState]    ? "normal" : "disabled"}]
        .m.edit entryconfigure 4 -state [expr {[$w selState]    ? "normal" : "disabled"}]
        .m.edit entryconfigure 5 -state [expr {$Clipboard != "" ? "normal" : "disabled"}]
    } else {
        .m.edit entryconfigure 0 -state disabled
        .m.edit entryconfigure 1 -state disabled
        .m.edit entryconfigure 3 -state disabled
        .m.edit entryconfigure 4 -state disabled
        .m.edit entryconfigure 5 -state disabled
    }
}

# ----------
# Window menue
# ----------

# com = command; -com = -command; -x = -xaccel
two_menu .m.win -tearoff false -postcommand winState
.m add cascade -menu .m.win -label "&Window"          -info "Modify windows"
.m.win add com -label "&Tile"    -com winTile    -x 4 -info "Tile windows to fill all space"
.m.win add com -label "&Cascade" -com winCascade -x 3 -info "Cascade windows"
.m.win add separator
.m.win add com -label "&Split"   -com winSplit   -x 2 -info "Split curreent window: show two views of same file"
.m.win add com -label "&Join"    -com winJoin    -x 1 -info "Join split windows"

# additional bindings for text
bind TwoText <Control-Insert> {event generate %W <Control-c>}
bind TwoText <Shift-Insert>   {event generate %W <Control-v>}
bind TwoText <Shift-Delete>   {event generate %W <Control-x>}
bind TwoText <Control-Delete> {event generate %W <Control-x>}

# set state of entries on menu open and append list of editors
proc winState {} {
    set w      [.mdi active].ed
    set slaves [.mdi slaves]
    set count  [llength $slaves]
    if {$w != ".ed"} {
        .m.win entryconfigure 3 -state [expr {![$w splitState] ? "normal" : "disabled"}]
        .m.win entryconfigure 4 -state [expr { [$w splitState] ? "normal" : "disabled"}]
    } else {
        .m.win entryconfigure 3 -state disabled
        .m.win entryconfigure 4 -state disabled
    }
    if {[.m.win index end] > 5} {
	.m.win delete 5 end
    }
    if {$count > 0} {
        .m.win add separator
	# add names of editors in reverse order
	set number 1
	incr count -1
	for {set idx $count} {$idx >= 0} {incr idx -1} {
	    set slave [lindex $slaves $idx]
	    set fname [$slave.ed filename]
	    if {$number < 10} {
		.m.win add command -command ".mdi show $slave" \
	            -label "&$number: [file tail $fname]" -info $fname
	    } else {
		.m.win add command -command ".mdi show $slave" \
	            -label "$number: [file tail $fname]" -info $fname
	    }
	    incr number
	}
        .m.win entryconfigure 0 -state normal
        .m.win entryconfigure 1 -state normal
    } else {
        .m.win entryconfigure 0 -state disabled
        .m.win entryconfigure 1 -state disabled
    }
}

# ----------
# Toolbar
# ----------

# b = button; -im = -image; -com = -command
grid [two_toolbar .tb] -columnspan 2 -sticky nsew
.tb entryconfigure 0 -image twoTearoff -tip "Tear"  -info "Tear off toolbar, make it floating"
.tb add b -im twoNew   -com fileNew    -tip "New"   -info "Create a new file"
.tb add b -im twoOpen  -com fileOpen   -tip "Open"  -info "Open a file"
.tb add b -im twoSave  -com fileSave   -tip "Save"  -info "Save current file" -off
.tb add b -im twoPrint -com filePrint  -tip "Print" -info "Print current file" -off
.tb add newline
.tb add b -im twoUndo  -com editUndo   -tip "Undo"  -info "Undo last changes" -off
.tb add b -im twoRedo  -com editRedo   -tip "Redo"  -info "Redo last undo" -off
.tb add separator
.tb add b -im twoCut   -com editCut    -tip "Cut"   -info "Cut selection and put it on the clipboard" -state disabled
.tb add b -im twoCopy  -com editCopy   -tip "Copy"  -info "Copy selection to the clipboard" -state disabled
.tb add b -im twoPaste -com editPaste  -tip "Paste" -info "Paste contents of clipboard" -state disabled
.tb add newline
.tb add check -label "B" -font [font create -family courier -size 10 -weight bold]  -padx 5
.tb add check -label "I" -font [font create -family courier -size 10 -slant italic] -padx 5
.tb add check -label "U" -font [font create -family courier -size 10 -underline 1]  -padx 5
.tb add combo -style windows -width 3 -alias combo
.tb add separator
.tb add b -im twoExit  -com fileExit  -tip "Exit"  -info "Exit programm"

.tb entrycommand combo add 10 12 14 16 18 20 24

# Switching to another editor will cause status change in toolbar
bind TwoMdiMaster <<ActiveOn>> {
    .tb entryconfigure 4 -state normal
    set w "[.mdi active].ed"
    .tb entryconfigure 3  -state [expr {[$w modState]  ? "normal" : "disabled"}]
    .tb entryconfigure 6  -state [expr {[$w undoState] ? "normal" : "disabled"}]
    .tb entryconfigure 7  -state [expr {[$w redoState] ? "normal" : "disabled"}]
    .tb entryconfigure 9  -state [expr {[$w selState]  ? "normal" : "disabled"}]
    .tb entryconfigure 10 -state [expr {[$w selState]  ? "normal" : "disabled"}]
}
bind TwoMdiMaster <<ActiveOff>> {
    .tb entryconfigure 4 -state disabled
    .tb entryconfigure 3  -state disabled
    .tb entryconfigure 6  -state disabled
    .tb entryconfigure 7  -state disabled
    .tb entryconfigure 9  -state disabled
    .tb entryconfigure 10 -state disabled
}

# status changes in toolbar according to action in editor
bind TwoEditor <<UndoOn>>  {.tb entryconfigure 6 -state normal   }
bind TwoEditor <<UndoOff>> {.tb entryconfigure 6 -state disabled }
bind TwoEditor <<RedoOn>>  {.tb entryconfigure 7 -state normal   }
bind TwoEditor <<RedoOff>> {.tb entryconfigure 7 -state disabled }
bind TwoEditor <<SelOn>> {
    .tb entryconfigure 9  -state normal
    .tb entryconfigure 10 -state normal
}
bind TwoEditor <<SelOff>> {
    .tb entryconfigure 9  -state disabled
    .tb entryconfigure 10 -state disabled
}
bind TwoEditor <<ModOn>> {
    .tb entryconfigure 3 -state normal
    setEditorTitle %W 1
}
bind TwoEditor <<ModOff>> {
    .tb entryconfigure 3 -state disabled
    setEditorTitle %W 0
}

# ----------
# Mdi window
# ----------

grid [two_mdi .mdi -width 600 -height 400 -slavebd 1 -style windows \
     -xscrollcommand ".sx set" -yscrollcommand ".sy set"] \
     [two_autosb .sy -command ".mdi yview"] \
     -sticky nsew
grid [two_autosb .sx -command ".mdi xview" -orient horizontal] \
     -sticky nsew
grid rowconfigure    . 1 -weight 1
grid columnconfigure . 0 -weight 1

if {![string match windows $tcl_platform(platform)]} {
    .mdi configure -font twoGui
}

# ----------
# Status line
# ----------

# Message from tooltips and current cursor position
# Caveat: setting width in position field -> only 3 digits are shown!
grid [frame .s -relief raised -bd 1] -columnspan 2 -stick nsew
pack [label .s.x -textvariable posX -anchor e -width 3 -relief sunken -padx 2] \
     -side right -padx 2 -pady 2
pack [label .s.xl -text "Col:" -anchor w -justify l -relief flat] \
     -side right -padx 2 -pady 2
pack [label .s.y -textvariable posY -anchor e -width 3 -relief sunken -padx 2] \
     -side right -padx 2 -pady 2
pack [label .s.yl -text "Row:" -anchor w -justify l -relief flat] \
     -side right -padx 2 -pady 2
pack [label .s.status -textvariable two_ttInfo -anchor w -justify l -relief sunken] \
     -fill x -expand true -side left -padx 2 -pady 2

if {$tcl_platform(platform) == "windows"} {
    .s configure -relief flat
}

# Switched to another editor
bind TwoText <FocusIn> {
    scan [%W index insert] "%%d.%%d" posY posX
    incr posX
}

bind TwoText <<PosChanged>> {
    set posX %x
    set posY %y
}

# ----------
# File Menue commands
# ----------

# Create a new slave widget.
# editor widget will create a new unique filename if filename is omitted.
# if filename is given: the editor will open and load this file
proc fileNew { {filename ""} } {
    set w [.mdi addSlave -closecmd onFileClose]
    pack [two_editor $w.ed -highlightthickness 0 -bg white -file $filename -style windows] \
         -expand 1 -fill both
    setEditorTitle $w.ed 0
    focus $w
}

# Open a file. If no fname is given: use dialog box.
# If already opened: raise editor widget. Add to history 
proc fileOpen { {fname ""} } {
    if {$fname == ""} {
	set fname [tk_getOpenFile]
    }
    if {$fname != ""} {
        # already opened?
	set opened 0
	foreach slave [.mdi slaves] {
	    if {[$slave.ed filename] == $fname} {
	        focus $slave
		set opened 1
		break
	    }
	}
	# open file
	if !$opened {
            fileNew $fname
	}
	addToFileHistory $fname
    }
}

# Save file. If filename is "New #" then use fileSaveAs
proc fileSave { {w ""} } {
    if {$w == ""} {
	withActive {
	    set w %W
	}
    }
    if [string match "New *" [$w filename]] {
        return [fileSaveAs $w]
    } else {
        $w save
    }
    return 1
}

# Save file with new name. Append this name to the history.
proc fileSaveAs { {w ""} } {
    if {$w == ""} {
	withActive {
	    set w %W
	}
    }
    set cmd tk_getSaveFile
    set pathname [$w filename]
    if {$pathname != ""} {
	set initialfile [file tail $pathname]
	lappend cmd -initialfile $initialfile
	if {$initialfile != $pathname} {
            lappend cmd -initialdir [file dirname $pathname]
	}
    }
    set fname [eval $cmd]
    if {$fname != ""} {
        $w save $fname
	addToFileHistory $fname
	.mdi entryconfigure [winfo parent $w] -title [file tail $fname]
	return 1
    }
    return 0
}

# Close editor. onFileClose will be called to do safety checks.
proc fileClose {} {
    withActive {
        set s [winfo parent %W]
        set m [winfo parent [winfo parent [winfo parent $s]]]
        $m delete $s
    }
}

# TODO
proc filePrint {} {
    withActive {
	puts "not yet"
    }
}

# Exit program. Close all editor first to give user a chance to save hist files.
proc fileExit {} {
    foreach slave [.mdi slaves] {
        if ![.mdi delete $slave] {
	    # break if user didn't close one of the editors
	    return
	}
    }
    exit
}

# ----------
# Edit Menue commands
# ----------

proc editUndo {} {
    withActive {
	%W undo
    }
}

proc editRedo {} {
    withActive {
	%W redo
    }
}

# Cut: use application's internal Clipboard
proc editCut {} {
    withActive {
        global Clipboard
	set Clipboard [%W get sel.first sel.last]
	%W delete sel.first sel.last
	# paste now enabled?
	# caveat: index number
	if {$Clipboard != ""} {
	    .tb entryconfigure 11 -state normal
	} else {
	    .tb entryconfigure 11 -state disabled
	}
        event generate %W <<SelState>> -sendevent 0
    }
}

# Copy: use application's internal Clipboard
proc editCopy {} {
    withActive {
        global Clipboard
	set Clipboard [%W get sel.first sel.last]
	# paste now enabled?
	# caveat: index number
	if {$Clipboard != ""} {
	    .tb entryconfigure 11 -state normal
	} else {
	    .tb entryconfigure 11 -state disabled
	}
    }
}

# Paste: use application's internal Clipboard
proc editPaste {} {
    withActive {
        global Clipboard
        if {$Clipboard != ""} {
            %W insert insert $Clipboard
	}
    }
}

# ----------
# Option Menue commands
# ----------

# turn X-Scrollbar on/off in all editor widgets
proc optScrollX {on} {
    foreach slave [.mdi slaves] {
        if $on {
	    grid $slave.ed.1.sx -row 1 -sticky nsew
	    grid $slave.ed.2.sx -row 1 -sticky nsew
	    $slave.ed.1.txt configure -wrap none
	    $slave.ed.2.txt configure -wrap none
        } else {
	    grid forget $slave.ed.1.sx
	    grid forget $slave.ed.2.sx
	    $slave.ed.1.txt configure -wrap char
	    $slave.ed.2.txt configure -wrap char
	}
    }
}

# ----------
# Window Menue commands
# ----------

proc winTile {} {
    .mdi tile
}

proc winCascade {} {
    .mdi cascade
}

proc winSplit {} {
    withActive {
        %W configure -split 1
    }
}

proc winJoin {} {
    withActive {
        %W configure -split 0
    }
}

# ----------
# Helper functions
# ----------

# evaluate command with "%W" substitution
# if there is a currently active editor (mdi slave)
proc withActive {command} {
    set w [.mdi active]
    if {$w != ""} {
        regsub -all {%W} $command "$w.ed" command
	uplevel $command
    }
}

# called if editor is about to be closed: safety checks
proc onFileClose {w} {
    if [$w.ed modState] {
        set fname [file tail [$w.ed filename]]
	switch [
	    tk_messageBox -message "File \"$fname\" is changed. Save it before closeing?" \
	       -type yesnocancel -default yes
	] {
	    yes {
	        if [fileSave $w.ed] {
		    return yes
		} else {
		    return no
		}
	    }
	    no {
	        return yes
	    }
	    default {
	        return no
	    }
	}
    }
    return yes
}

# append to file history, but only once
proc addToFileHistory {fname} {
    global Recent

    # append to history, but only once
    set idx [lsearch -exact $Recent $fname]
    if {$idx >= 0} {
	set Recent [lreplace $Recent $idx $idx]
    }
    set Recent [linsert $Recent 0 $fname]
    if {[llength $Recent] > 8} {
	set Recent [lrange $Recent 0 7]
    }
}

# set title (mdi!) of editor widget
proc setEditorTitle {w modified} {
    if {!$modified} {
        .mdi entryconfigure [winfo parent $w] -title "[file tail [$w filename]]"
    } else {
        .mdi entryconfigure [winfo parent $w] -title "[file tail [$w filename]] *"
    }
}

# test
bind TwoText <Control-Key-5> {optScrollX 0}
bind TwoText <Control-Key-6> {optScrollX 1}
