# html.tcl --
#
#	HTML Widget. This does not create a toplevel, it just provides
#	the hypertext widget to be used e.g. within a help system.
#	This widget uses a subset of HTML 2.0 :-)
#
# Copyright (c) 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/html.tcl,v 1.8 1999/11/03 09:16:09 fjl Exp $

package provide twoHtml 1.0
package require twoUtil
package require twoTooltip
package require http

# This might interfere with other parts trying to configure the http
# useragent - but it is a shortcoming of the http library itself to
# make configurations interpreter-wide:-(
::http::config -useragent "TWO simple html widget 1.0"

# Create fonts (catch as they might already exist)
catch {font create twoHtmlH1   -family times     -size 18 -weight bold}
catch {font create twoHtmlH2   -family times     -size 14 -weight bold}
catch {font create twoHtmlH3   -family times     -size 12 -weight bold -slant italic}
catch {font create twoHtmlH4   -family times     -size 12 -weight bold}
catch {font create twoHtmlH5   -family times     -size 12 -slant italic}
catch {font create twoHtmlH6   -family times     -size 12}
catch {font create twoHtmlP    -family helvetica -size 12}
catch {font create twoHtmlPB   -family helvetica -size 12 -weight bold}
catch {font create twoHtmlPI   -family helvetica -size 12              -slant italic}
catch {font create twoHtmlPBI  -family helvetica -size 12 -weight bold -slant italic}
catch {font create twoHtmlPC   -family courier   -size 12}
catch {font create twoHtmlPBC  -family courier   -size 12 -weight bold}
catch {font create twoHtmlPIC  -family courier   -size 12              -slant italic}
catch {font create twoHtmlPBIC -family courier   -size 12 -weight bold -slant italic}
catch {font create twoHtmlS    -family times     -size 1}

namespace eval ::two::html {
    namespace export two_html
    variable Html
    variable Visited
    variable Preload
    
    namespace import ::two::util::two_parseArgs
    
    # variable Html
    #
    # $w,hist -	history as tcl list (elements are a list of URL and Title)
    # $w,hcurrent -	index of current page in the history
    # $w,cbase -	current base url for relative urls
    # $w,home -	home page
    # $w,nonce -    nonce to give each link tag a new name
    # $w,active -   active tag (needed to properly handle button press on hyperlinks)
    # $w,activefg - normal foreground color of active tag
    # $w,h,<p> -	html text for htmlPath <p> (pre-set)
    # $w,images -	names of all temprary images which are included in the widget
    # $w,ltags -	names of all temprary tags used for links
    # $w,cursor -	Original cursor while dragging with control-B2
    # mouseMoved -	Non-zero means the mouse has moved a significant
    #		amount since the button went down (so, for example,
    #		start dragging out a selection).
    # x, y -	Last known mouse coordinates for scanning and auto-scanning.
    
    # variable Visited
    #
    # hashtable that holds with all keys which to already visited html pages.
    
    # variable Preload
    #
    # hashtable that holds with all pages which are pre-set (not to be
    # taken from URL).
}

image create bitmap TwoHtmlBullet -data {
\#define bullet_width 10
\#define bullet_height 10
static char bullet_bits[] = {
   0x78, 0x00, 0xfc, 0x00, 0xe6, 0x01, 0xc3, 0x03, 0xc3, 0x03, 0xe7, 0x03,
   0xff, 0x03, 0xfe, 0x01, 0xfc, 0x00, 0x78, 0x00};
}

image create bitmap TwoHtmlBrokenImage -data {
\#define broken_width 26
\#define broken_height 32
static unsigned char broken_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0xfe, 0xff, 0x01, 0x00, 0xfe, 0xff, 0x03, 0x00,
   0x06, 0x80, 0x07, 0x00, 0x06, 0x80, 0x0f, 0x00, 0x06, 0x80, 0x1d, 0x00,
   0x86, 0x87, 0x3b, 0x00, 0xc6, 0x8c, 0x75, 0x00, 0xc6, 0x8c, 0xeb, 0x00,
   0x06, 0x86, 0xd5, 0x01, 0x06, 0x83, 0xab, 0x01, 0x06, 0x83, 0xff, 0x01,
   0x06, 0x80, 0xff, 0x01, 0x06, 0x03, 0x80, 0x01, 0x06, 0x03, 0x80, 0x01,
   0x06, 0x00, 0x80, 0x01, 0x06, 0x00, 0x80, 0x01, 0x86, 0x87, 0x87, 0x01,
   0xc6, 0xcc, 0x8c, 0x01, 0xc6, 0xcc, 0x8c, 0x01, 0x06, 0x06, 0x86, 0x01,
   0x06, 0x03, 0x83, 0x01, 0x06, 0x03, 0x83, 0x01, 0x06, 0x00, 0x80, 0x01,
   0x06, 0x03, 0x83, 0x01, 0x06, 0x03, 0x83, 0x01, 0x06, 0x00, 0x80, 0x01,
   0x06, 0x00, 0x80, 0x01, 0x06, 0x00, 0x80, 0x01, 0xfe, 0xff, 0xff, 0x01,
   0xfe, 0xff, 0xff, 0x01, 0x00, 0x00, 0x00, 0x00};
}

# ::two::html::two_html --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".html"); the original widget procedure
#	is renamed to "_two$w" (like "_two.html") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::html::two_html {w args} {
    variable Html

    # Initialise the globals
    array set Html "
        $w,home		index.html
        $w,linkcolor	blue2
        $w,alinkcolor	red
        $w,vlinkcolor	purple4
        $w,hist		{}
        $w,cbase	{}
        $w,hcurrent	0
        $w,nonce	0
        $w,active	{}
        $w,images	{}
        $w,ltags	{}
    "
    # See if this option is for us...
    two_parseArgs args Html($w,home)       "-home"
    two_parseArgs args Html($w,linkcolor)  "-linkcolor"
    two_parseArgs args Html($w,alinkcolor) "-alinkcolor"
    two_parseArgs args Html($w,vlinkcolor) "-vlinkcolor"
    set Html($w,cbase) $Html($w,home)

    # Create text widget
    uplevel [list text $w -wrap word -insertwidth 0] $args
#    uplevel text $w -wrap word $args
    
    # only needed for grabs to really go nowhere:-)
    frame $w.grab

    # Only use our own bindings
    bindtags $w [list $w TwoHtml [winfo toplevel $w] all]
#    bindtags $w [list $w Text [winfo toplevel $w] all]
    
    # Configure tags
    # SP: Spacing between paragraphs
    # S: Small font
    # HR: Horizontal rule
    # H1, H2, H3: Heading (Level 1-6)
    # LI1, LI2: List elements (Level 1-6)
    # DT1, DT2: Definition Term (Level 1-6)
    # DD1, DD2: Definition Description (Level 1-6)
    # DC: DT and DC in one line for compact definition lists
    # CEN: Center
    # REF: Hyper Reference (only the base tag, more a created on demand)
    # CUR: Navigaton cursor frame (for tab navigation)
    # P, PB, PI, PBI, PC, PBC, PIC, PBIC: normal Paragraphs (P) with
    #     Bold (B), Italics (I) and/or Courier (C) fonts
    $w tag configure SP  -spacing1 10
    $w tag configure HR  -font twoHtmlS  -background black -bgstipple gray25

    $w tag configure H1  -font twoHtmlH1 -justify center -spacing1 30 -spacing3 30
    $w tag configure H2  -font twoHtmlH2 -spacing1 20 -spacing3 20
    $w tag configure H3  -font twoHtmlH3 -spacing1 20 -spacing3 20
    $w tag configure H4  -font twoHtmlH4 -spacing1 20
    $w tag configure H5  -font twoHtmlH5 -spacing1 20
    $w tag configure H6  -font twoHtmlH6 -spacing1 20

#    $w tag configure UL
#    $w tag configure OL
#    $w tag configure DL

    $w tag configure LI1 -tabs {0.2i center 0.4i}                -lmargin2 0.4i
    $w tag configure LI2 -tabs {0.6i center 0.8i} -lmargin1 0.4i -lmargin2 0.8i
    $w tag configure LI3 -tabs {1.0i center 1.2i} -lmargin1 0.8i -lmargin2 1.2i
    $w tag configure LI4 -tabs {1.4i center 1.6i} -lmargin1 1.2i -lmargin2 1.6i
    $w tag configure LI5 -tabs {1.8i center 2.0i} -lmargin1 1.6i -lmargin2 2.0i
    $w tag configure LI6 -tabs {2.2i center 2.4i} -lmargin1 2.0i -lmargin2 2.4i

    $w tag configure DC1 -tabs {0.4i}                -lmargin2 0.4i
    $w tag configure DC2 -tabs {0.8i} -lmargin1 0.4i -lmargin2 0.8i
    $w tag configure DC3 -tabs {1.2i} -lmargin1 0.8i -lmargin2 1.2i
    $w tag configure DC4 -tabs {1.6i} -lmargin1 1.2i -lmargin2 1.6i
    $w tag configure DC5 -tabs {2.0i} -lmargin1 1.6i -lmargin2 2.0i
    $w tag configure DC6 -tabs {2.4i} -lmargin1 2.0i -lmargin2 2.4i

#    $w tag configure DT1
    $w tag configure DT2 -lmargin1 0.4i -lmargin2 0.4i
    $w tag configure DT3 -lmargin1 0.8i -lmargin2 0.8i
    $w tag configure DT4 -lmargin1 1.2i -lmargin2 1.2i
    $w tag configure DT5 -lmargin1 1.6i -lmargin2 1.6i
    $w tag configure DT6 -lmargin1 2.0i -lmargin2 2.0i

    $w tag configure DD1 -lmargin1 0.4i -lmargin2 0.4i
    $w tag configure DD2 -lmargin1 0.8i -lmargin2 0.8i
    $w tag configure DD3 -lmargin1 1.2i -lmargin2 1.2i
    $w tag configure DD4 -lmargin1 1.6i -lmargin2 1.6i
    $w tag configure DD5 -lmargin1 2.0i -lmargin2 2.0i
    $w tag configure DD6 -lmargin1 2.4i -lmargin2 2.4i
    
    $w tag configure PRE -wrap none
    $w tag configure CEN -justify center
    $w tag configure REF -underline true
    $w tag configure CUR -borderwidth 2 -relief solid

    foreach fnt {S P PB PI PBI PC PBC PIC PBIC} {
        $w tag configure $fnt -font twoHtml$fnt
    }
    
    # common bindings for all REF tags
    $w tag bind REF <Enter> [namespace code {_two%W configure -cursor hand2}]
    $w tag bind REF <Leave> [namespace code {_two%W configure -cursor left_ptr}]

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "

    # catch widget destruction in order to clean up Html
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::html::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::html::Widget {w opt args} {
    variable Html
    variable Preload

    # See if this option is for us...
    switch -exact -- $opt {
	configure {
	    set retval ""
	    two_parseArgs args Html($w,home)       "-home"
	    two_parseArgs args Html($w,linkcolor)  "-linkcolor"
	    two_parseArgs args Html($w,alinkcolor) "-alinkcolor"
	    two_parseArgs args Html($w,vlinkcolor) "-vlinkcolor"
	    if {![string length $Html($w,cbase)]} {
	        set Html($w,cbase) $Html($w,home)
            }

            if {[llength $args]} {
		set retval [eval [list _two$w $opt] $args]
	    }
	    if {[llength $Html($w,hist)]} {
                ShowPage $w [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0] 1
	    }
            return $retval
	}
	cget {
	    switch -exact -- $args {
	        -home       { return $Html($w,home)       }
	        -linkcolor  { return $Html($w,linkcolor)  }
	        -alinkcolor { return $Html($w,alinkcolor) }
	        -vlinkcolor { return $Html($w,vlinkcolor) }
		default     { return [eval [list _two$w $opt] $args] }
	    }
	}
	home {
	    return [ShowPage $w $Html($w,home) 0 1]
	}
	show {
	    if {[llength $args] != 1} {
	    	return -code error "wrong # args: should be \"$w show page\""
	    }
            return [ShowPage $w [lindex $args 0]]
	}
	history -
	hist {
	    set hist ""
	    foreach link $Html($w,hist) {
	        set hist [linsert $hist 0 $link]
	    }
	    return $hist
	}
	current {
	    return [expr {[llength $Html($w,hist)] - $Html($w,hcurrent)}]
	}
	goto {
	    set retval ""
	    if {[llength $args] != 1} {
	    	return -code error "wrong # args: should be \"$w goto index\""
	    }
	    set pos [lindex $args 0]
	    if {$pos <= 0} {
	        return -code error "invalid index \"$pos\""
	    }
	    if {$pos > [llength $Html($w,hist)]} {
	        return -code error "invalid index \"$pos\""
	    }
	    set pos [expr {[llength $Html($w,hist)] - $pos}]
	    
	    if {$Html($w,hcurrent) != $pos} {
                if {$Html($w,hcurrent) == [llength $Html($w,hist)] - 1} {
                    event generate $w <<NextOn>>
                }
                if {$Html($w,hcurrent) == 0} {
                    event generate $w <<PrevOn>>
                }

                set Html($w,hcurrent) $pos
                set retval [ShowPage $w \
                    [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0] 1
                ]

                if {$Html($w,hcurrent) == [llength $Html($w,hist)] - 1} {
                    event generate $w <<NextOff>>
                }
                if {$Html($w,hcurrent) == 0} {
                    event generate $w <<PrevOff>>
                }
	    }
	    return $retval
	}
	nextState {
	    if {$Html($w,hcurrent) < [llength $Html($w,hist)] - 1} {
	        return 1
	    } else {
		return 0
	    }
	}
	next {
	    set retval ""
	    if {[llength $args] > 0} {
	    	return -code error "wrong # args: should be \"$w next\""
	    }
	    if {$Html($w,hcurrent) == 0} {
	        event generate $w <<PrevOn>>
	    }
	    if {$Html($w,hcurrent) < [llength $Html($w,hist)] - 1} {
	        incr Html($w,hcurrent)
	        set retval [ShowPage $w \
	            [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0] 1
	        ]
	    }
	    if {$Html($w,hcurrent) == [llength $Html($w,hist)] - 1} {
	        event generate $w <<NextOff>>
	    }
	    return $retval
	}
	prevState {
	    if {$Html($w,hcurrent) > 0 } {
	        return 1
	    } else {
		return 0
	    }
	}
	prev {
	    set retval ""
	    if {[llength $args] > 1} {
	    	return -code error "wrong # args: should be \"$w prev ?steps?\""
	    }
	    if {$Html($w,hcurrent) == [llength $Html($w,hist)] - 1} {
	        event generate $w <<NextOn>>
	    }
	    if {![string length $args]} {
	        set step 1
	    } else {
	        set step [lindex $args 0]
	    }
	    if {$Html($w,hcurrent) > $step - 1} {
	        incr Html($w,hcurrent) "-$step"
	        set retval [ShowPage $w \
                    [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0] 1
	        ]
	    }
	    if {$Html($w,hcurrent) == 0} {
	        event generate $w <<PrevOff>>
	    }
	    return $retval
	}
	refresh {
	    return [ShowPage $w \
                [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0] 1
	    ]
	}
	add {
	    set url [lindex $args 0]
	    set Preload($url) [lindex $args 1]
	}
	append {
	    append Preload([lindex $args 0]) [lindex $args 1]
	}
	default {
            return [eval [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::html::_Destroy --
#
#	Call this procedure when the html widget is destroyed.
#	Will clean widget-specific Elements of Html and destroy
#	the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::html::_Destroy {w} {
    variable Html

    # clear Html
    foreach key [array names Html "$w,*"] {
        unset Html($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoHtml-class bindings
#-------------------------------------------------------------------------

bind TwoHtml <1>     {focus %W}

bind TwoHtml <Control-2> {
    if {!$tk_strictMotif} {
        set ::two::html::Html(%W,cursor) [::two::html::_two%W cget -cursor]
        ::two::html::_two%W configure -cursor fleur
        ::two::html::_two%W scan mark %x %y
        set ::two::html::Html(x) %x
        set ::two::html::Html(y) %y
        set ::two::html::Html(mouseMoved) 0
    }
}
bind TwoHtml <Control-B2-Motion> {
    if {!$tk_strictMotif} {
        if {(%x != $::two::html::Html(x)) || (%y != $::two::html::Html(y))} {
            set ::two::html::Html(mouseMoved) 1
        }
        if {$::two::html::Html(mouseMoved)} {
            ::two::html::_two%W scan dragto %x %y
        }
    }
}
bind TwoHtml <Control-ButtonRelease-2> {
    if {!$tk_strictMotif} {
        ::two::html::_two%W configure -cursor $::two::html::Html(%W,cursor)
    }
}

bind TwoHtml <Up>    {%W yview scroll -1 unit}
bind TwoHtml <Down>  {%W yview scroll  1 unit}
bind TwoHtml <Prior> {%W yview scroll -1 page}
bind TwoHtml <Next>  {%W yview scroll  1 page}

bind TwoHtml <Return> {
    ::two::html::SelectShow %W
}
bind TwoHtml <space> {
    ::two::html::SelectShow %W
}
bind TwoHtml <Left> {
    ::two::html::SelectPrev %W
}
bind TwoHtml <Right> {
    ::two::html::SelectNext %W
}
bind TwoHtml <Tab> {
    ::two::html::SelectNext %W
    # don't change focus
    break
}
bind TwoHtml <Shift-Tab> {
    ::two::html::SelectPrev %W
    # don't change focus
    break
}
bind TwoHtml <Control-Tab> {
    # Needed only to keep <Tab> binding from triggering;  doesn't
    # have to actually do anything.
}
bind TwoHtml <Shift-Control-Tab> {
    # Needed only to keep <Tab> binding from triggering;  doesn't
    # have to actually do anything.
}
bind TwoHtml <FocusIn> {
    ::two::html::_two%W tag configure CUR -relief solid
}
bind TwoHtml <FocusOut> {
    ::two::html::_two%W tag configure CUR -relief flat
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::html::MakeAbsUrl --
#
#	Takes an url and tries to make an absoulte url from it. If it is
#	relative, then it will be taken relative to the current base url.
#
# Arguments:
#	w		Path to html widget
#	varName		Name of variable containing the url
#
# Sideeffects:
#	The Variable $varame is possibly modified to an absolute url

proc ::two::html::MakeAbsUrl {w varname} {
    upvar $varname url
    variable Html

    # Only file: and http: allowed, and both are 5 chars:-)
    set urlProto [string tolower [string range $url 0 4]]
    if {[string match "http:" $urlProto]} {
        # nothing to do
    } elseif {[string match "file:" $urlProto]} {
        # using a protocol specifier should guarantee absoute filenames.
        # No further processing needed, just remove the "file:" part
        set url [string range $url 5 end]
    } else {
        # o.k., url could be relative. Now, what is the reference?
        regexp {^[^\#]*} $Html($w,cbase) ref
        # Only if the url has any chars should it be relative, else
        # take the ref (the previous location) directly
        if {[string length $url]} {
            # get the protocol descriptor of the reference file
            set refProto [string tolower [string range $ref 0 4]]
            # is the reference an html file?
            if {[string match "http:" $refProto]} {
                # extract base from url
                regexp -nocase {^(http://[^/]+)(/.*)?$} $ref trash base add
                if {[string match "/*" $url]} {
                    set url "$base$url"
                } elseif {[string match "#*" $url]} {
                    set url "$ref$url"
                } else {
                    if {![string match "*/" $add]} {
                        regsub {/+[^/]+$} $add {} add
                    }
                    append base $add
                    set url "$base/$url"
                }
            } else {
                # no html, so assume file. Remove the "file:" Protocol
                # selector (only if present)
                if {[string match "file:" $refProto]} {
                    set ref [string range $ref 5 end]
                }
                regexp {^[^\#]*} $ref ref

                if {[string match "#*" $url]} {
                    set url "$ref$url"
                } else {
                    # the file command handles absoulte vs. relative file names
                    # properly...
                    set url [file join [file dirname $ref] $url]
                }

                # ... but leaves the ".." in the filename. This would make the
                # filename grow forever and make matching filenames against
                # the built-in list of files impossilbel.
                set fname ""
                if {[string match "absolute" [file pathtype $ref]]} {
                    set top /
                } else {
                    set top [file split [pwd]]
                }

                foreach part [file split $url] {
                    switch -- $part {
                        . {
                        }
                        .. {
                            if {![llength $fname]} {
                                set fname $top
                                set top /
                            }
                            set fname [lreplace $fname end end]
                        }
                        default {
                            lappend fname $part
                        }
                    }
                }
                if {[llength $fname]} {
                    set url [eval file join $fname]
                } else {
                    set url [file nativename /]
                }
            }
        } else {
            set url $ref
        }
    }
}

# ::two::html::ShowPage --
#
#	show page accessed by "url" in the html widget "w". Also handles
#	the history: pages are appended to the history (this is skipped
#	if the fromhist parameter is set).
#
# Arguments:
#	w		Path to html widget
#	url		Name of html page which should be shown
#	fromhistory	true if page is accessed through history
#	isabsolute	true if this page is meant as absoulte url (home...)
#
# Returns:
#	Empty string if successfull, or the error message

proc ::two::html::ShowPage {w url {fromhistory 0} {isabsolute 0}} {
    variable Html
    variable History
    variable Visited
    global TwoHtmlTitle
    global two_ttInfo

    #
    # Step 1: turn relative url to absolute url
    #

    # extract anchor part
    regexp "(\[^\#\]*)\#?(.*)" $url trash url anchor

    # only navigate within current page?
    if {![string length $url] && [string length $anchor]} {
        if {[catch {_two$w yview "A_$anchor"} err]} {
            set two_ttInfo $err
            return $err
        } else {
            # current base url...
            regexp {^[^\#]*} $Html($w,cbase) url
            append url \# $anchor
            set Visited($w,$url) ""
            return ""
        }
    }

    if {!$isabsolute && !$fromhistory} {
        MakeAbsUrl $w url
    }

    #
    # Step 2: load and display html code
    #

    grab $w.grab
    after idle [namespace code [list _two$w configure -cursor watch]]
    
    set html ""

    # is this in fact just an image?
    switch -glob -- $url {
        *.gif -
        *.jpg -
        *.jpeg -
        *.xbm -
        *.xpm -
        *.png {
            set html "<HTML><HEAD>
              <TITLE>$url</TITLE>
              </HEAD><BODY>
              <IMG SRC=\"$url\">
              </BODY></HTML>"
        }
        default {
            if {[catch {set html [GetUrl $w url]} err]} {
                set two_ttInfo $err
                _two$w configure -cursor left_ptr
                grab release $w.grab
                two_mb -type ok -default ok -icon error -title "HTML Widget" \
                    -message $err -parent $w
                return $err
            }
        }
    }
    set Html($w,cbase) $url
    set title [RenderHtml $w $html]

    _two$w configure -cursor left_ptr
    grab release $w.grab

    # show requested anchor
    if {[string length $anchor]} {
        catch {_two$w yview "A_$anchor"}
    }
    set TwoHtmlTitle $title
    set Visited($w,$url) ""
    
    #
    # Step 3: add to history
    #

    if {!$fromhistory} {
        if {![llength $Html($w,hist)]} {
            set Html($w,hist) [list [list $url $title]]
        } elseif {![string match $url \
            [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0]
        ]} {
            if {$Html($w,hcurrent) == 0} {
                event generate $w <<PrevOn>>
            }
            if {$Html($w,hcurrent) <= [llength $Html($w,hist)]} {
                set Html($w,hist) [lrange $Html($w,hist) 0 $Html($w,hcurrent)]
                incr Html($w,hcurrent)
                event generate $w <<NextOff>>
            }
            lappend Html($w,hist) [list $url $title]
        }
    }

    set two_ttInfo ""
    return ""
}

# ::two::html::GetUrl --
#
#	Retrieve a file through the url. The url must be handled over
#	by reference and not by value, as a redirect might well
#	change it.
#
# Arguments:
#	w		Path to html widget
#	urlvar		Name variable with Url to retrieve
#
# Returns:
#	Contents of that url

proc ::two::html::GetUrl {w urlvar} {
    variable Html
    variable Preload
    global two_ttInfo
    upvar $urlvar url

    set content ""

    if {[info exists Preload($url)]} {
        # Fine, this page is already pre-loaded...
        set content $Preload($url)
    } else {
        # try to retrieve the page and render it
        # or return with an error message
        if {[string match "http:" [string tolower [string range $url 0 4]]]} {
            while {1} {
                # http... set timeout to 8 sec.
                # might throw an error... fine
                set token [::http::geturl $url -timeout 8000]
                # redirect?
                array set meta [set ${token}(meta)]
                if {[info exists meta(Location)]} {
                    # string trim due to bug in http package
                    set url [string trimleft $meta(Location)]
                    unset $token
                    unset meta
                } else {
                    break
                }
            }
            set content [::http::data $token]
            unset $token
        } else {
            # might throw an error... fine, this will be reported
            # to the calling function which should handle it.
            set content [two_readFile -binary $url]
        }
    }
    set two_ttInfo $url
    update idletasks

    # this should be "return $content", but this is faster!?
    set content
}

# ::two::html::RenderHtml --
#
#	Render and display an html page (and return the title).
#
# Arguments:
#	w		Path to html widget
#	html		Name of html page which should be shown
#
# Returns:
#	The Title of the HTML file (if it is set), else an empty
#	string

proc ::two::html::RenderHtml {w html} {
    variable Html
    variable Visited

    # clean up
    _two$w delete 1.0 end
    set Html($w,nonce) 0
    if {[llength $Html($w,images)]} {
        eval image delete $Html($w,images)
        set Html($w,images) {}
    }
    if {[llength $Html($w,ltags)]} {
        two_tooltip delete $w
        eval [list _two$w tag delete] $Html($w,ltags)
        set Html($w,ltags) {}
    }

    # removing comments would be easy as of tcl8.1, due to the
    # non greedy .*?, but trying to be compatible to tcl8.0 makes
    # this unavailable: <!--.*?-->
    set rest $html
    set html ""
    while {[set idx [string first "<!--" $rest]] >= 0} {
        append html [string range $rest 0 [expr "$idx -1"]]
        set rest [string range $rest [expr "$idx +4"] end]
        if {[set idx [string first "-->" $rest]] >= 0} {
            set rest [string range $rest [expr "$idx +3"] end]
        } else {
            append html $rest
            set rest ""
        }
    }
    append html $rest

    # TODO: embedded TCL <SCRIPT NAME="TwoTcl">code</SCRIPT>
    # would be quite nice:-)

    # Split into parts: the first one is pure text, all subsequent parts are
    # a tag and the text up to the next tag or to the end of the file
    set parts [split $html "<"]

    # the first part is NOT a tag; insert it directly if it ain't whitespace
    set part [string trim [lindex $parts 0]]
    if {[string length $part]} {
        _two$w insert end $part P
    }

    # state info:
    set Para	""	;# current paragraph format (list, filled as stack!)
    set Char	""	;# character format (B=Bold, I=Italics, C=Courier)
    set Link	""	;# currently within an anchor? (contains the tag to add)
    set Name	""	;# Name of anchor which should be added
    set Indent	""	;# additinal paragraph format for indendation
    set Level	0	;# indendation level
    set Show	1	;# 1 if text is to be shown, else 0
    set Center1	0	;# true if next paragraph is top be centred
    set Center2	0	;# true if centering untill next /CENTER
    # OL($Level) used for current position within ordered lists

    # only used once at start of next paragraph:
    set Break	0	;# Linebreak: 0 = none, 1 = if needed, 2 = forced
    set Space	0	;# if true, then the next paragraph need to have spacing
    set Tags	"P"	;# tags from previous line
    
    # The <TITLE> of the HTML file
    set Title ""

    # Work through the parts
    foreach part [lrange $parts 1 end] {
    	# Split that part into tag and remaining txt
    	set tmp [split $part ">"]
    	set tag [lindex $tmp 0]
    	set txt [join [lrange $tmp 1 end] ">"]
    	set img ""

    	# normalize the tag and split it into tag and options
    	set tmp [split [string trim $tag] " "]
    	set tag [string toupper [lindex $tmp 0]]
    	set opt [join [lrange $tmp 1 end] " "]
    	
    	# make list of option-value-pairs from opt
    	set opt [SplitParam $opt]

        # current column is:
        set Col [lindex [split [_two$w index "end-1c"] "."] 1]
        
        #
        # Step 1: analyse tags
        #
        
        switch -exact -- $tag {
            HEAD {
                set Show 0
            }
            /HEAD {
                set Show 1
            }
            TITLE {
                set Title [string trim $txt]
            }
            BASE {
                set base [HasOpt $opt HREF]
                # The MakeBasUrl proc expects the base url to be a real URL.
                # So turn the base into a real url - with the proper base:-)
                append base "/dummy.html"
                if {[string length $base]} {
                    set Html($w,cbase) $base
                }
            }
            BR {
                incr Break 2
            }
            P {
                incr Break 2
                set Space 1
                set Center1 [HasOpt $opt ALIGN CENTER]
            }
            /P {
                incr Break
                set Space 1
                set Center1 0
            }
            H1 -
            H2 -
            H3 -
            H4 -
            H5 -
            H6 {
                incr Break
                lappend Para $tag
                set Center [HasOpt $opt ALIGN CENTER]
            }
            /H1 -
            /H2 -
            /H3 -
            /H4 -
            /H5 -
            /H6 {
                incr Break
                set Para [lreplace $Para end end]
            }
            HR {
    	        if {$Col} {
                    if {[string match " " [_two$w get "insert -1c" insert]]} {
                        _two$w delete "insert -1c" insert
                    }
    	            _two$w insert end "\n" $Tags
    	        }
                _two$w insert end "\n" S
                if {$Space} {
                    _two$w tag add "SP" "insert linestart"
                    set Space 0
                }
                set Tags {S HR}
                _two$w insert end " " $Tags
                set Col 1
                incr Break 2
                set Space 1
            }
            PRE {
                incr Break
                set Space 1
                lappend Para $tag
                append Char "C"
            }
            /PRE {
                incr Break
                set Space 1
                set Para [lreplace $Para end end]
                regsub "C" $Char "" Char
                if {[string match "\n" [_two$w get "insert -1c" insert]] &&
                    [string match PRE [lindex [_two$w tag names "insert -1c"] end]]
                } {
                    _two$w delete "insert -1c" insert
                }
            }
            CENTER {
                incr Break
                incr Center2
            }
            /CENTER {
                incr Break
                incr Center2 -1
                if {$Center2 < 0} {set Center2 0}
            }
            UL -
            DIR -
            MENU -
            BLOCKQUOTE {
                incr Break
                set Space 1
                incr Level
                if {$Level > 6} {
                    set Indent DD6
                } else {
                    set Indent DD$Level
                }
                lappend Para UL
            }
            OL {
                incr Break
                set Space 1
                incr Level
                set OL($Level) [HasOpt $opt START]
                if {![string length $OL($Level)] || [catch {expr $OL($Level)}]} {
                    set OL($Level) 1
                }
                lappend Para $tag
                if {$Level > 6} {
                    set Indent DD6
                } else {
                    set Indent DD$Level
                }
            }
            DL {
                incr Break
                set Space 1
                incr Level
                if {[string match "COMPACT" [string toupper [lindex $opt 0]]]} {
                    lappend Para DLC
                    if {$Level > 6} {
                        set Indent LI6
                    } else {
                        set Indent LI$Level
                    }
                } else {
                    lappend Para $tag
                    if {$Level > 6} {
                        set Indent DD6
                    } else {
                        set Indent DD$Level
                    }
                }
            }
            /UL -
            /DIR -
            /MENU -
            /OL -
            /DL -
            /BLOCKQUOTE {
                incr Break
                set Space 1
                incr Level -1
                if {$Level < 0} {set Level 0}
                set Para [lreplace $Para end end]
                if {$Level > 6} {
                    set Indent DD6
                } elseif {$Indent > 0} {
                    set Indent DD$Level
                }
            }
            LI {
                set Indent LI$Level
                # Can't use the "incr Break" as we insert additional text
                # right here, but the Break would occur afterwards...
                set Break 0
    	        if {$Col && [string match " " [_two$w get "insert -1c" insert]]} {
                    _two$w delete "insert -1c" insert
                    incr Col -1
                }
                if {$Col} {
                    _two$w insert end "\n" $Tags
    	        }
                set txt [string trimleft $txt]
                if {$Level > 6} {
                    set li LI6
                } else {
                    set li LI$Level
                }
                if {[string match OL [lindex $Para end]]} {
                    _two$w insert end "\t$OL($Level).\t" "$li P"
                    incr OL($Level)
                } else {
                    _two$w insert end "\t" "$li P"
                    _two$w image create end -image TwoHtmlBullet
                    _two$w tag add "$li P" "insert -1c" insert
                    _two$w insert end "\t" "$li P"
                }
                set Col 3
                if {$Space} {
                    _two$w tag add "SP" "insert linestart"
                    set Space 0
                }
                if {$Level > 6} {
                    set Indent DD6
                } else {
                    set Indent DD$Level
                }
            }
            DT {
                if {[string match "DLC" [lindex $Para end]]} {
                    set Break 0
                    if {$Col && [string match " " [_two$w get "insert -1c" insert]]} {
                        _two$w delete "insert -1c" insert
                        incr Col -1
                    }
                    if {$Col} {
                        _two$w insert end "\n" $Tags
                    }
                    if {$Level > 6} {
                        set Indent LI6
                    } else {
                        set Indent LI$Level
                    }
                    _two$w insert end "\t" "$Indent P"
                    if {$Space} {
                        _two$w tag add "SP" "insert linestart"
                        set Space 0
                    }
                } else {
                    incr Break
                    if {$Level > 6} {
                        set Indent DT6
                    } else {
                        set Indent DT$Level
                    }
                }
            }
            DD {
                if {[string match "DLC" [lindex $Para end]]} {
                    if {$Level > 6} {
                        set Indent LI6
                    } else {
                        set Indent LI$Level
                    }
                    if {[string match " " [_two$w get "insert -1c" insert]]} {
                        _two$w delete "insert -1c" insert
                    }
                    _two$w insert end "\t" "$Indent P"
                    if {$Space} {
                        _two$w tag add "SP" "insert linestart"
                        set Space 0
                    }
                } else {
                    incr Break
                    if {$Level > 6} {
                        set Indent DD6
                    } else {
                        set Indent DD$Level
                    }
                }
            }
            IMG {
                set img [HasOpt $opt SRC]
                set alt [HasOpt $opt ALT]
                if {[string first "&" $alt] >= 0} {
                    set alt [DecodeCharEntities $alt]
                }
            }
            A {
                set Name [HasOpt $opt NAME]
                set href [HasOpt $opt HREF]
                set abs $href
                MakeAbsUrl $w abs
                if {[string length $href]} {
                    set Link link$Html($w,nonce)
                    incr Html($w,nonce)
                    lappend Html($w,ltags) $Link
    
                    if {![info exists Visited($w,$abs)]} {
                        _two$w tag configure $Link -foreground $Html($w,linkcolor)
                    } else {
                        _two$w tag configure $Link -foreground $Html($w,vlinkcolor)
                    }
                    _two$w tag bind $Link <1> [namespace code "
                        set Html($w,active) $Link
                        set Html($w,activefg) \[_two$w tag cget $Link -foreground\]
                        _two$w tag configure $Link -foreground $Html($w,alinkcolor)
                    "]
                    _two$w tag bind $Link <ButtonRelease-1> [namespace code "
                        _two$w tag configure $Link -foreground \$Html($w,activefg)
                        if {\[lsearch -exact \[_two$w tag names @%x,%y\] \$Html($w,active)\] != -1} {
                            $w show $href
                        }
                    "]
                    two_tooltip $w -tag $Link -info $href
                }
            }
            /A {
                set Link ""
            }
            B -
            STRONG {
                append Char "B"
            }
            /B -
            /STRONG {
                regsub "B" $Char "" Char
            }
            I -
            EM -
            CITE -
            VAR {
                append Char "I"
            }
            /I -
            /EM -
            /CITE -
            /VAR {
                regsub "I" $Char "" Char
            }
            ADDRESS {
                incr Break
                append Char "I"
            }
            /ADDRESS {
                incr Break
                regsub "I" $Char "" Char
            }
            TT -
            CODE -
            KBD -
            SAMP {
                append Char "C"
            }
            /TT -
            /CODE -
            /KBD -
            /SAMP {
                regsub "C" $Char "" Char
            }
    	}
    	
    	#
    	# Step 2: add the text snippets to the widget
    	#

    	if {$Show} {
    	    # current paragraph format
            set p [lindex $Para end]

            # remove newlines and collapse multiple spaces
            # if this text is not marked as preformatted
            if {![string match "PRE" $p]} {
                regsub -all "\[\n\t \]+" $txt { } txt
            }

            # cut off any leading whitespace at beginning of a new line
            # (in preforatted text only remove leading newlines)
            if {$Break} {
                # remove trailing space before newline in order to make centered
                # text really centered...
                if {$Col && [string match " " [_two$w get "insert -1c" insert]]} {
                    _two$w delete "insert -1c" insert
                }
                if {[string match "PRE" $p]} {
                    set txt [string trimleft $txt "\n"]
                } else {
                    set txt [string trimleft $txt]
                }
            }

            # Forced line break: add newline in any case
    	    if {$Break >= 2} {
                if {$Col} {
                    _two$w insert end "\n" $Tags
                } else {
                    _two$w insert end "\n" "P"
                }
                set Break 0
            }
            
            # Format and insert the current piece of text
            if {[string length $txt] || [string length $img]} {
                # text should start at a new line, but cursor currently
                # is in another column
                if {$Break && $Col} {
                    _two$w insert end "\n" $Tags
                }
                set Break 0
                
                # paragraph format
                if {[string length $p]} {
                    set Tags [list $p]
                } else {
                    set Tags ""
                }
                if {$Center1 || $Center2} {
                    lappend Tags CEN
                }
                if {[string length $Indent]} {
                    lappend Tags $Indent
                }
                
                # character format
                if {![string match "H?" $p]} {
                    set c "P"
                    if {[string match "*B*" $Char]} {append c "B"}
                    if {[string match "*I*" $Char]} {append c "I"}
                    if {[string match "*C*" $Char]} {append c "C"}
                    lappend Tags $c
                }
                if {[string length $Link]} {
                    lappend Tags REF $Link
                }
                if {[string length $Name]} {
                    _two$w mark set "A_$Name" insert
                    _two$w mark gravity "A_$Name" left
                    set Name ""
                }

                # insert an image (before text as image is no container tag
                # and thus might well be followed by some text)
                if {[string length $img]} {
                    AddImage $w $img
                    foreach t $Tags {
                        _two$w tag add $t "insert -1c" insert
                    }
                    if {$Space} {
                        set Space 0
                        _two$w tag add "SP" "insert -1c" insert
                    }
                    set imgtag img$Html($w,nonce)
                    incr Html($w,nonce)
                    lappend Html($w,ltags) $imgtag
                    _two$w tag add $imgtag "insert -1c" insert
                    if {[string length $Link]} {
                        two_tooltip $w -tag $imgtag -tip $alt -info $href
                    } else {
                        two_tooltip $w -tag $imgtag -tip $alt -info $img
                    }
                }

                # really insert the text
                if {[string length $txt]} {
                    # transform some character classes; only if there are any:-)
                    if {[string first "&" $txt] >= 0} {
                        set txt [DecodeCharEntities $txt]
                    }
                    if {$Space} {
                        set Space 0
                        set pos [_two$w index insert]
                        _two$w insert end $txt $Tags
                        _two$w tag add "SP" $pos "$pos + 1c"
                    } else {
                        _two$w insert end $txt $Tags
                    }
                }
            }
        }
    }
    return $Title
}

# ::two::html::AddImage --
#
#	Add an image to the Html widget
#
# Arguments:
#	w		Path to html widget
#	img		Path to image file

proc ::two::html::AddImage {w img} {
    variable Html

    # special "protocol" for already existing images
    if {[string match "icon:*" [string tolower $img]]} {
        if {[catch {
            _two$w image create end -image [string range $img 5 end]
        }]} {
            _two$w image create end -image TwoHtmlBrokenImage
        }
    } else {
        MakeAbsUrl $w img
        set imgData ""
        catch {set imgData [GetUrl $w img]}

        # try to build the image
        if {[string length $imgData] &&
            (
            ![catch {set name [image create bitmap -data $imgData]}] ||
            ![catch {set name [image create photo  -data $imgData]}] ||
            ![catch {set name [image create photo  -data [base64_encode $imgData]]}]
            )
        } {
            _two$w image create end -image $name
            lappend Html($w,images) $name
        } else {
            _two$w image create end -image TwoHtmlBrokenImage -align center
        }
    }
}

# ::two::html::SelectNext --
#
#	Set html page internal Focus to next link in the page.
#
# Arguments:
#	w		Path to html widget

proc ::two::html::SelectNext {w} {
    variable Html

    # any links in here?
    if {[llength [_two$w tag ranges REF]]} {
        # Where are we now?
        if {[llength [_two$w tag ranges CUR]]} {
            set curBeg [_two$w index CUR.first]
            set curEnd [_two$w index CUR.last]
        } else {
            set curBeg 1.0
            set curEnd 1.0
        }
        # look for next position
        set next [_two$w tag nextrange REF $curEnd end]
        if {![llength $next]} {
            set next [_two$w tag nextrange REF 1.0 $curBeg]
        }
        # found any? Select this link
        if {[llength $next]} {
            _two$w tag remove CUR 1.0 end
            _two$w tag add CUR [lindex $next 0] [lindex $next 1]
            _two$w see [lindex $next 0]
        }
    }
}

# ::two::html::SelectPrev --
#
#	Set html page internal Focus to previous link in the page.
#
# Arguments:
#	w		Path to html widget

proc ::two::html::SelectPrev {w} {
    variable Html

    # any links in here?
    if {[llength [_two$w tag ranges REF]]} {
        # Where are we now?
        if {[llength [_two$w tag ranges CUR]]} {
            set curBeg [_two$w index CUR.first]
            set curEnd [_two$w index CUR.last]
        } else {
            set curBeg 1.0
            set curEnd 1.0
        }
        # look for previous position
        set prev [_two$w tag prevrange REF $curBeg 1.0]
        if {![llength $prev]} {
            set prev [_two$w tag prevrange REF end $curEnd]
        }
        # found any? Select this link
        if {[llength $prev]} {
            _two$w tag remove CUR 1.0 end
            _two$w tag add CUR [lindex $prev 0] [lindex $prev 1]
            _two$w see [lindex $prev 0]
        }
    }
}

# ::two::html::SelectShow --
#
#	Show page whose link has been selected.
#
# Arguments:
#	w		Path to html widget

proc ::two::html::SelectShow {w} {
    variable Html

    # anything selected?
    if {[llength [_two$w tag ranges CUR]]} {
        # get position
        set pos [_two$w bbox CUR.first]
        foreach {xmin ymin width height} $pos {break}
        set x [expr {$xmin + $width  / 2}]
        set y [expr {$ymin + $height / 2}]
        # create the same events as if the mouse pointer were used
        event generate $w <Enter>           -x $x -y $y
        event generate $w <ButtonPress-1>   -x $x -y $y
        event generate $w <ButtonRelease-1> -x $x -y $y
    }
}

# ::two::html::SplitParam --
#
#	Split the parameters of a tag into a paired list
#	(option, value). Options without values are handled
#	as if the value were empty.
#
# Arguments:
#	param		Parameter string from tag
#
# Returns:
#	A single list with an even number of elements. Elements are
#	pairs of option and value: {opt1} {val1} {opt2} {val2}...

proc ::two::html::SplitParam {param} {
    set opt ""
    set param [string trim $param]

    while {[string length $param]} {
        # get the next key
        regexp {^([^= ]*)( *=? *)(.*)$} $param trash key sep val
        set key [string trim $key]
        if {[string length $key]} {
            set key [string toupper $key]
            if {[string match "*=*" $sep]} {
                # separator given, so we have to extract the value
                set val [string trim $val]
                if {[string match "\"*" $val]} {
                    # quoted... so look for closing quote
                    # TODO: right now, masked quotes (backslash!) are not
                    # handled properly: they act as normal end-quotes
                    regexp {^\"([^\"]*)\"?(.*)$} $val trash val param
                } else {
                    # no quotes, so just take next word
                    regexp {^([^ ]*) ?(.*)$} $val trash val param
                }
                lappend opt $key $val
            } else {
                # no separator, so just take this key and continue
                set param $val
                lappend opt $key ""
            }
            set param [string trim $param]
        } else {
            # no key, so we're finished
            set param ""
        }
    }
    return $opt
}

# ::two::html::HasOpt --
#
#	Query a list of option-value-pairs if a specified option is
#	set and return that value. Or, if an additional value is
#	given to the command, reuturn true if this option is set
#	and false otherwise.
#
# Arguments:
#	opt		key value pairs like those from SplitParam
#	searchKey	the key to search for
#	args		optional; the value which we look for
#
# Returns:
#	Two parameters given: the value to this key or, if no such
#	key is available, an empty string.
#	Three parameters given: True if key-value pair is found,
#	else False.

proc ::two::html::HasOpt {opt searchKey args} {
    if {[llength $args]} {
        set searchVal [string toupper [join $args]]
        set matchVal 1
        set retval 0
    } else {
        set matchVal 0
        set retval ""
    }
    set searchKey [string toupper $searchKey]

    foreach {key val} $opt {
        if {[string compare [string toupper $key] $searchKey]==0} {
            if {$matchVal} {
                if {[string compare [string toupper $val] $searchVal]==0} {
                    set retval 1
                } else {
                    set retval 0
                }
            } else {
                set retval $val
            }
        }
    }
    return $retval
}

# ::two::html::DecodeCharEntities --
#
#	Decode character entities to ISO latin-1 (e.g. &auml; to )
#
# Arguments:
#
#	encoded -		the encoded text
#
# Retruns:
#	Decoded text

proc ::two::html::DecodeCharEntities {encoded} {
    variable CharEntities

    set txt ""
    foreach part [split $encoded "&"] {
        set l [string first ";" $part]
        set char [string range $part 0 [expr {$l - 1}]]
        if {[string match {\#[0-9]*} $char]} {
            append txt [format %c [string range $char 1 end]]
        } elseif {[info exists CharEntities($char)]} {
            append txt $CharEntities($char)
        } elseif {[string length $char]} {
            append txt "&" $char ";"
        }
        append txt [string range $part [incr l] end]
    }
    # could be "return $txt" as well:-)
    set txt
}

variable ::two::html::CharEntities
array set ::two::html::CharEntities {
    lt     "<"
    gt     ">"
    amp    "&"
    quot   \"
    nbsp   " "

    pound  ""
    yen    ""
    brvbar ""
    sect   ""
    copy   ""
    laquo  ""
    not    ""
    shy    ""
    reg    ""
    deg    ""
    plusmn ""
    sup1   ""
    sup2   ""
    sup3   ""
    acute  ""
    micro  ""
    para   ""
    middot ""
    raquo  ""
    frac14 ""
    frac12 ""
    frac34 ""
    iquest ""

    auml   ""
    ouml   ""
    uuml   ""
    Auml   ""
    Ouml   ""
    Uuml   ""
    szlig  ""

    aelig  ""
    AElig  ""

    aacute ""
    acirc  ""
    agrave ""
    eacute ""
    ecirc  ""
    egrave ""
    oslash ""
    oacute ""
    ocirc  ""
    ograve ""
    uacute ""
    ucirc  ""
    ugrave ""
}


# ::two::html::base64_encode --
#
# 	do base-64 encoding (see RFC 2045 for details). This version
#	works on strings; it transform the input string given as
#	argument into the base64-encoded version which is returned.
#	CAVEAT EMPTOR: When reading the input data from whatever
#	source (like a file), make sure that you don't use any kind
#	of EOL translation => fconfigure $fh -translation binary
#
# Arguments:
# 	orig -		binary data which is to be encoded
#
# Results:
#	base64-encoded string

proc ::two::html::base64_encode {orig} {

    set base64en {
        A B C D E F G H I J K L M N O P Q R S T U V W X Y Z
	a b c d e f g h i j k l m n o p q r s t u v w x y z
	0 1 2 3 4 5 6 7 8 9 + /
    }
    set pos 0
    set out ""

    foreach {i0 i1 i2} [split $orig {}] {
        set in "$i0$i1$i2"
        set num [string length $in]
	binary scan \000$in\000\000 I b32
	for {set idx 18} {$idx >= (3-$num)*6} {incr idx -6} {
	    append out [lindex $base64en [expr {($b32>>$idx) & 0x3F}]]
	}
	if {$num < 3} {
	    for {set idx 0} {$idx < 3-$num} {incr idx} {
	        append out "="
	    }
	    append out "\n"
	    set pos 0
	} elseif {$pos == 17} {
	    append out "\n"
	    set pos 0
	} else {
	    incr pos
	}
    }
    return $out
}

namespace import ::two::html::*
