# keyrec.tcl --
#
#	Record key press events. Usefull e.g. when building new dialog
#	widgets which are in need of calls to update, but don't want
#	to loose key events.
#
# Copyright (c) 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/keyrec.tcl,v 1.2 1999/10/26 13:12:18 fjl Exp $

package provide twoKeyrec 1.0

namespace eval ::two::keyrec {
    namespace export two_keyrec
    variable Keyrec
    
    # variable Keyrec
    #
    # w -		Path to recorder widget
    # focus -	Old focus prior to starting the recorder
    # rec -		True if currently recording
    # e -		Events recorded so far
    
    array set Keyrec {
        e	""
    }
}

# ::two::keyrec::two_keyrec --
#
#	Record events
#
# Arguments:
#	cmd		Sub-Command
#	args		arguments and options

proc ::two::keyrec::two_keyrec {cmd args} {
    variable Keyrec
    
    switch $cmd {
        init {
            if {[info exists Keyrec(w)] && [winfo exists $Keyrec(w)]} {
                destroy $Keyrec(w)
            }
            if {[llength $args]>1} {
                return -code error "Worng # arguments: should be \"two_keyrec init ?widget?\""
            } elseif  {[llength $args]==1} {
                set Keyrec(w) [lindex $args 0]
            } else {
                set Keyrec(w) .__two_keyrec
            }
            frame $Keyrec(w) -class TwoKeyrec -takefocus 0
            place $Keyrec(w) -x -10 -y -10 -width 1 -height 1
        }
        start {
            if {![info exists Keyrec(w)]} {
                two_keyrec init
            }
            set Keyrec(focus) [focus]
            set Keyrec(rec)   1
            set Keyrec(e)     ""
            focus -force $Keyrec(w)
            update idletasks
        }
        stop {
            if {![info exists Keyrec(rec)] || !$Keyrec(rec)} {
                return -code error "Must call \"two_keyrec start\" first"
            }
            update
            if {[winfo exists $Keyrec(focus)]} {
                focus -force $Keyrec(focus)
            } else {
                focus -force .
            }
            set Keyrec(rec) 0
        }
        replay {
            if {[llength $args] > 1} {
                return -code error "Wrong # arguments: should be \"two_keyrec replay ?widget?\""
            } elseif {[llength $args] == 1} {
                set w [lindex $args 0]
            } else {
                set w [list [focus]]
            }
            if {![string length $w] || ![winfo exists $w]} {
                set w .
            }
            foreach e $Keyrec(e) {
                eval [list event generate $w] $e
            }
        }
    }
    return
}

proc ::two::keyrec::addBindings {} {
    foreach sequence {
        <KeyPress> <KeyRelease>
        <Shift-KeyPress> <Shift-KeyRelease>
        <Control-KeyPress> <Control-KeyRelease>
        <Shift-Control-KeyPress> <Shift-Control-KeyRelease>
    } {
        bind TwoKeyrec $sequence "
            if \$::two::keyrec::Keyrec(rec) {
                lappend ::two::keyrec::Keyrec(e) \[list $sequence -keycode %k -keysym %K -time %t\]
            }
            break
        "
    }

    bind TwoKeyrec <FocusOut> {
        if {$::two::keyrec::Keyrec(rec) && [string length [focus]]} {
            set ::two::keyrec::Keyrec(focus) [focus]
            focus -force $::two::keyrec::Keyrec(w)
        }
    }
}
::two::keyrec::addBindings

namespace import ::two::keyrec::*
