# mdi.tcl --
#
#	MDI Manager widget. This widget can hold an arbitrary number
#	of child widgets which look like KDE or Win95. The user can
#	manipulate them in the usual way.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/mdi.tcl,v 1.72 1999/10/28 16:01:36 fjl Exp $

package provide twoMdi 1.0
package require twoUtil
package require twoMenu

# Workaround: Tk needs these procedures for focus traversion
# but won't load it form inside the namespace (bug in tcl8.0 up to p1)
catch {auto_load tk_focusNext}
catch {auto_load tk_focusPrev}

#-------------------------------------------------------------------------
# naming convention:
# $m, master -	Main MDI widget, contains slaves.
# $b, border -	Direct child of master, for internal use only. Provides
#		border for slave widget. Caveat: [winfo slaves $m] will
#		return the borders (and popup menus), not the slaves.
# $s, slave -	The tcl/tk frame widget which is returned to the MDI
#		user. More details: see doc for AddSlave procedure
#-------------------------------------------------------------------------

option add *TwoMdiMaster.background                grey51
option add *TwoMdi.relief                          raised
option add *TwoMdi.cursor                          left_ptr
option add *TwoMdiTitle.Button.highlightThickness  0
option add *TwoMdiTitle.Button.background          grey77
#option add *MdiTitle.Button.activeBackground       grey77

if {[string match "unix" $tcl_platform(platform)]} {
    option add *MdiMaster.borderWidth		0
} else {
    option add *MdiMaster.relief		sunken
    option add *MdiMaster.borderWidth		2
}

namespace eval ::two::mdi {
    namespace export two_mdi
    variable Mdi
    
    namespace import ::two::util::two_parseArgs
    
    # variable Mdi
    #
    # $w,slavebg -	background color of slave window
    # $w,slavebd -	border width of slave window
    # $w,font -	font of title of slave window
    # $w,titlebg -	background color of inactive title
    # $w,titlefg -	foreground color of inactive title
    # $w,activebg -	background color of active title
    # $w,activefg -	foreground color of active title
    # nonce -	nonce to give any toplevel a new number
    # doubleclick -	time (in ms) for double clicks (used in slave's system menu)
    # offset -	current offset for new slaves
    # offset0 -	start of offset
    # step -	relative offset for next slave
    # slaveW -	minimum initial width of new slaves
    # slaveH -	minimum initial height of new slaves
    # borderW -	width of slave window's border
    # Drag -	1 if drag is going on
    # Cursor -	name of current cursor; used to determine which edge is drag'ed
    # Master -	current master
    # Master,w -	width of current master
    # Master,h -	height of current master
    # Master,x1 -	left x of current master
    # Master,y1 -	left y of current master
    # Master,x2 -	right x of current master
    # Master,y2 -	right y of current master
    # Border -	path to border of current slave window
    # x -		distance from right border at start of drag
    # y -		distance from bottom border at start of drag
    # xx -		helper value for x dragging of window
    # yy -		helper value for y dragging of window
    # $m,style -	style of mdi (windows or unix)
    # $m,icon -	prefix of icons according to current style
    # $m,arrangeicons -	if 1, then icons will be moved on resize
    # $m,xscrollcommand -	command to use on x scroll
    # $m,yscrollcommand -	command to use on y scroll
    # $m,slaves -	list of all slaves (in arbitrary order)
    # $m,xOffset -	current x offset of slave's bounding box to the left of the master
    # $m,yOffset -	current y offset of slave's bounding box to the top of the master
    # $m,maxWidth -	width of bounding box of all slaves and the master himself
    # $m,maxHeight -	height of bounding box of all slaves and the master himself
    # $m,icons -	list of positions of all minimized slaves of $m
    # $m,max -	1 if all slaves of master $m are maximized
    # $m,active -	path name of currently active slave of master $m
    # $m,resizeQ -	Id of an after idle script for resizing (proc QueueResize)
    # $b,min -	1 if slave windget is minimized (
    # $b,place -	records "place info" for normal slave
    # $b,icon -	position of icon (index, not coordinates)
    # $b,title -	string to display as title
    # $b,x -	x position of slave's border
    # $b,y -	y position of slave's border
    # $b,lastmenu -	time of last invocation of slave menu (to detect double-B1)
    
    array set Mdi {
        nonce	0
        doubleclick	500
        offset	10
        offset0	10
        step        20
        slaveW      300
        slaveH      200
        borderW	3
        Drag	0
        x		0
        y		0
    }
}

#-------------------------------------------------------------------------
# Bitmaps
#-------------------------------------------------------------------------

# window handles, Windows Look
image create bitmap mdiWinMin -foreground black -data {
    \#define winmin_width 13
    \#define winmin_height 11
    static unsigned char winmin_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0xf8, 0x01, 0xf8, 0x01, 0x00, 0x00, };
}
image create bitmap mdiWinMax -foreground black -data {
    \#define winmax_width 13
    \#define winmax_height 11
    static unsigned char winmax_bits[] = {
      0x00, 0x00, 0xfc, 0x07, 0xfc, 0x07, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04,
      0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0xfc, 0x07, 0x00, 0x00, };
}
image create bitmap mdiWinRest -foreground black -data {
    \#define winrest_width 13
    \#define winrest_height 11
    static unsigned char winrest_bits[] = {
      0x00, 0x00, 0xf0, 0x03, 0xf0, 0x03, 0x10, 0x02, 0xfc, 0x02, 0xfc, 0x02,
      0x84, 0x03, 0x84, 0x00, 0x84, 0x00, 0xfc, 0x00, 0x00, 0x00, };
}
image create bitmap mdiWinClose -foreground black -data {
    \#define winclose_width 13
    \#define winclose_height 11
    static unsigned char winclose_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x18, 0x06, 0x30, 0x03, 0xe0, 0x01, 0xc0, 0x00,
      0xe0, 0x01, 0x30, 0x03, 0x18, 0x06, 0x00, 0x00, 0x00, 0x00, };
}

# window handles, KDE Look
image create bitmap mdiKdeMin -foreground white -background grey44 -data {
    \#define kdemin_main_width 16
    \#define kdemin_main_height 16
    static unsigned char kdemin_main_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x80, 0x03, 0x80, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
} -maskdata {
    \#define kdemin_mask_width 16
    \#define kdemin_mask_height 16
    static unsigned char kdemin_mask_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x80, 0x03, 0x80, 0x02, 0x80, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
}
image create bitmap mdiKdeMax -foreground white -background grey44 -data {
    \#define kdemax_main_width 16
    \#define kdemax_main_height 16
    static unsigned char kdemax_main_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0xfc, 0x1f, 0x04, 0x00, 0x04, 0x08, 0x04, 0x08,
      0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0xf4, 0x0f,
      0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
} -maskdata {
    \#define kdemax_mask_width 16
    \#define kdemax_mask_height 16
    static unsigned char kdemax_mask_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0xfc, 0x1f, 0xfc, 0x1f, 0x0c, 0x18, 0x0c, 0x18,
      0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0xfc, 0x1f,
      0xfc, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
}
image create bitmap mdiKdeRest -foreground white -background grey44 -data {
    \#define kderest_main_width 16
    \#define kderest_main_height 16
    static unsigned char kderest_main_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf8, 0x1f, 0x08, 0x10, 0x08, 0x10,
      0x08, 0x10, 0x08, 0x10, 0x08, 0x10, 0x08, 0x10, 0x08, 0x10, 0x08, 0x10,
      0xf8, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
} -maskdata {
    \#define kderest_mask_width 16
    \#define kderest_mask_height 16
    static unsigned char kderest_mask_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0xfc, 0x1f, 0xfc, 0x1f, 0x0c, 0x18, 0x0c, 0x18,
      0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0xfc, 0x1f,
      0xfc, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
}
image create bitmap mdiKdeClose -foreground white -background grey44 -data {
    \#define kdeclose_main_width 16
    \#define kdeclose_main_height 16
    static unsigned char kdeclose_main_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x04, 0x10, 0x04, 0x08, 0x08, 0x04, 0x10, 0x02,
      0x20, 0x01, 0x40, 0x00, 0x40, 0x00, 0x20, 0x01, 0x10, 0x02, 0x08, 0x04,
      0x04, 0x08, 0x04, 0x10, 0x00, 0x00, 0x00, 0x00, };
} -maskdata {
    \#define kdeclose_mask_width 16
    \#define kdeclose_mask_height 16
    static unsigned char kdeclose_mask_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x0c, 0x30, 0x1c, 0x38, 0x38, 0x1c, 0x70, 0x0e,
      0xe0, 0x07, 0xc0, 0x03, 0xc0, 0x03, 0xe0, 0x07, 0x70, 0x0e, 0x38, 0x1c,
      0x1c, 0x38, 0x0c, 0x30, 0x00, 0x00, 0x00, 0x00, };
}

# text icon (shown as upper left handle in window titlebar)
image create bitmap mdiText -foreground white -data {
\#define text_width 11
\#define text_height 13
static unsigned char text_bits[] = {
  0x7f, 0x00, 0xc1, 0x00, 0x41, 0x01, 0x59, 0x02, 0xc1, 0x07, 0x01, 0x04,
  0xf9, 0x04, 0x01, 0x04, 0x01, 0x04, 0xf9, 0x04, 0x01, 0x04, 0x01, 0x04,
  0xff, 0x07, };
}

#-------------------------------------------------------------------------
# procedures
#-------------------------------------------------------------------------

# ::two::mdi::two_mdi --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.win"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.win") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	m		Path to master widget
#	args		Widget arguments and options

proc ::two::mdi::two_mdi {m args} {
    variable Mdi
    global tcl_platform

    incr Mdi(nonce)

    # This is a multi-style widget
    set ::two::SetStyleCmds($m) [list $m configure -style]

    # Initialise the globals
    array set Mdi "
	$m,slavebg   grey77
	$m,slavebd   2
	$m,font      {}
	$m,titlebg   grey51
	$m,titlefg   grey77
	$m,activebg  blue4
	$m,activefg  white
	$m,max       0
	$m,icons     {}
	$m,active    {}
        $m,style     $tcl_platform(platform)
	$m,arrangeicons   1
	$m,xscrollcommand {}
	$m,yscrollcommand {}
	$m,slaves    {}
    "

    # See if this option is for us...
    two_parseArgs args Mdi($m,slavebg)   "-slavebg"
    two_parseArgs args Mdi($m,slavebd)   "-slavebd"
    two_parseArgs args Mdi($m,font)      "-font"
    two_parseArgs args Mdi($m,titlebg)   "-titlebg"
    two_parseArgs args Mdi($m,titlefg)   "-titlefg"
    two_parseArgs args Mdi($m,activebg)  "-activebg"
    two_parseArgs args Mdi($m,activefg)  "-activefg"
    two_parseArgs args Mdi($m,style)     "-style"
    two_parseArgs args Mdi($m,arrangeicons)   "-arrangeicons"
    two_parseArgs args Mdi($m,xscrollcommand) "-xscrollcommand"
    two_parseArgs args Mdi($m,yscrollcommand) "-yscrollcommand"
    
    if {[string match "unix" $Mdi($m,style)]} {
        set Mdi($m,icon) "mdiKde"
    } else {
        set Mdi($m,icon) "mdiWin"
    }

    # Create master widget
    uplevel [list frame $m] $args -class TwoMdiMaster -takefocus 0
    
    # Create popup menu for window icon. All slaves windows use this menu,
    # there is no need for separate "system" menus for each slave.
    two_menu $m.menu -tearoff 0
    $m.menu add command -label "Restore" -alias rest -state disabled \
        -command [namespace code {Restore $Mdi(Border)}]
    $m.menu add command -label "Minimize" -alias min \
        -command [namespace code {Minimize $Mdi(Border)}]
    $m.menu add command -label "Maximize" -alias max \
        -command [namespace code {Maximize $Mdi(Border)}]
    $m.menu add command -label "Lower" -alias lower \
        -command [namespace code {LowerSlave $Mdi(Border)}]
    $m.menu add separator
    $m.menu add command -label "Close" -alias close \
        -command [namespace code {DeleteSlave $Mdi(Border)}]

    # Create root popup menu
    two_menu $m.root -tearoff 0
    $m.root add command -label "Shuffle Up" -alias up -state disabled \
        -command [namespace code "Shuffle up $m"]
    $m.root add command -label "Shuffle Down" -alias down -state disabled \
        -command [namespace code "Shuffle down $m"]

    # Use our own bindings
    bindtags $m [list $m TwoMdi TwoMdiMaster [winfo toplevel $m] all]

    # Move widget command
    rename $m _two$m

    # Install our replacement command
    namespace eval :: "
        proc $m {opt args} {
	    uplevel [namespace code Widget] $m \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Mdi
    bind $m <Destroy> [namespace code "_Destroy $m"]

    # Return the correct result...
    return $m
}

# ::two::mdi::Widget --
#
#	replacement for widget command
#
# Arguments:
#	m		Path to real master widget
#	opt		Widget command...
#	arg		optional args

proc ::two::mdi::Widget {m opt args} {
    variable Mdi

    # See if this option is for us...
    switch -exact -- $opt {
	configure {
	    set retval ""
	    two_parseArgs args Mdi($m,slavebg)   "-slavebg"
	    two_parseArgs args Mdi($m,slavebd)   "-slavebd"
	    two_parseArgs args Mdi($m,font)      "-font"
	    two_parseArgs args Mdi($m,titlebg)   "-titlebg"
	    two_parseArgs args Mdi($m,titlefg)   "-titlefg"
	    two_parseArgs args Mdi($m,activebg)  "-activebg"
	    two_parseArgs args Mdi($m,activefg)  "-activefg"
	    set st [two_parseArgs args Mdi($m,style)          "-style"]
	    set ai [two_parseArgs args Mdi($m,arrangeicons)   "-arrangeicons"]
	    set rx [two_parseArgs args Mdi($m,xscrollcommand) "-xscrollcommand"]
	    set ry [two_parseArgs args Mdi($m,yscrollcommand) "-yscrollcommand"]
	    if {$ai || $rx || $ry} {
	        Resize $m
	    }
	    if {[llength $args]} {
		set retval [eval [list _two$m $opt] $args]
	    }
	    if {[string match "unix" $Mdi($m,style)]} {
		set Mdi($m,icon) "mdiKde"
	    } else {
		set Mdi($m,icon) "mdiWin"
	    }
	    foreach s $Mdi($m,slaves) {
		ConfigureSlave $s
		if {$st} {
		    SetSlaveStyle $s
		}
	    }
	    return $retval
	}
	cget {
	    switch -exact -- [lindex $args 0] {
	        -slavebg   { return $Mdi($m,slavebg)   }
	        -slavebd   { return $Mdi($m,slavebd)   }
	        -font      { return $Mdi($m,font)      }
	        -titlebg   { return $Mdi($m,titlebg)   }
	        -titlefg   { return $Mdi($m,titlefg)   }
	        -activebg  { return $Mdi($m,activebg)  }
	        -activefg  { return $Mdi($m,activefg)  }
	        -style     { return $Mdi($m,style)     }
	        -arrangeicons   { return $Mdi($m,arrangeicons)   }
		-xscrollcommand { return $Mdi($m,xscrollcommand) }
		-yscrollcommand { return $Mdi($m,yscrollcommand) }
		default   { return [eval [list _two$m $opt] $args] }
	    }
	}
	new -
	addSlave {
	    return [eval [list AddSlave $m] $args]
	}
	slaves {
	    set slaves ""
	    foreach slave $Mdi($m,slaves) {
	        lappend slaves $slave.w.s
	    }
	    return $slaves
	}
	entryconfigure {
	    set slave [lindex $args 0]
	    set b     [winfo parent [winfo parent $slave]]
	    set args  [lrange $args 1 end]

	    if {[two_parseArgs args Mdi($b,title) "-title"]} {
	        $b.w.t.label configure -text $Mdi($b,title)
	    }
	    two_parseArgs args Mdi($b,closecmd) "-closecmd"
	    if {[two_parseArgs args spec "-pos"]} {
	        SetSlavePos $m $b $spec
	    }
	    if {[llength $args]} {
	        eval [list $b.w.s configure] $args
	    }
	}
	entrycget {
	    if {[llength $args] != 2} {
	        return -code error "wrong # args: should be \"two_mdi entrycget slave option\""
	    }
	    set slave [lindex $args 0]
	    set b     [winfo parent [winfo parent $slave]]
	    set arg   [lindex $args 1]
	    switch -exact -- $arg {
	        -title     { return $Mdi($b,title)    }
	        -closecmd  { return $Mdi($b,closecmd) }
	        -pos       {
	            if $Mdi($b,min) {
	                set state min
	                set pos $Mdi($b,place)
	            } elseif $Mdi($m,max) {
	                set state max
	                set pos $Mdi($b,place)
	            } else {
                        set state normal
                        set pos [place info $b]
	            }
	            return [list $state $pos]
	        }
		default    { return [uplevel [list $slave cget] $arg] }
	    }
	}
	tile {
	    Tile $m
	}
	cascade {
	    Cascade $m
	}
	active {
	    # might not exist if deleted and no new focus set
	    if {[winfo exists $Mdi($m,active)]} {
		return $Mdi($m,active).w.s
	    } else {
	        return ""
	    }
	}
	delete {
	    if {[llength $args] > 1} {
	    	return -code error "wrong # args: should be \"two_mdi delete slave\""
	    }
	    # delete slave window
	    return [DeleteSlave [winfo parent [winfo parent [lindex $args 0]]]]
	}
	show {
	    if {[llength $args] > 1} {
	    	return -code error "wrong # args: should be \"two_mdi show slave\""
	    }
	    set b [winfo parent [winfo parent [lindex $args 0]]]
	    set m [winfo parent $b]
	    if {$Mdi($b,min)} {
		if {$Mdi($m,max)} {
	            Maximize $b
		} else {
	            RestoreMin $b
		}
	    } else {
		raise $b
                SetFocus $b.w.s
	    }
	}
	maximize {
	    # get active child
	    set slave [$m active]
	    if {[string length $slave]} {
	        Maximize [winfo parent [winfo parent $slave]]
	    }
	}
	minimize {
	    if {![llength $args]} {
                foreach b [AllSlaveBorders $m] {
		    Minimize $b
		}
	    } else {
		foreach slave $args {
	            Minimize [winfo parent [winfo parent $slave]]
		}
	    }
	}
	restore {
	    if {![llength $args]} {
                foreach b [AllSlaveBorders $m] {
		    Restore $b
		}
	    } else {
		foreach slave $args {
	            Restore [winfo parent [winfo parent $slave]]
		}
	    }
	}
	xview {
	    eval [list Xview $m] $args
	}
	yview {
	    eval [list Yview $m] $args
	}
	default {
            return [eval [list _two$m $opt] $args]
	}
    }
    return
}

# ::two::mdi::_Destroy --
#
#	Call this procedure when the Mdi widget is destroyed.
#	Will clean widget-specific Elements of Mdi and destroy
#	the original widget command.
#
# Arguments:
#	m		Path to master widget

proc ::two::mdi::_Destroy {m} {
    variable Mdi

    unset ::two::SetStyleCmds($m)

    # clear Mdi
    foreach key [array names Mdi "$m,*"] {
        unset Mdi($key)
    }
    # remove the original widget command
    rename $m ""
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoMdi-class bindings
#-------------------------------------------------------------------------

# ---
# Bindings for master widget
# ---

# any mouse button gives root popup
# use some displacement from mouse coordinates to avoid problems
# with grab when clicking in fast sequence
bind TwoMdi <ButtonPress> {
    tk_popup %W.root %X [expr {%Y + 1}]
}

# resize master will lead to recalculation of scrollbars
bind TwoMdi <Configure> {
    ::two::mdi::Resize %W
}

# ---
# Bindings for button in unix-style slave widget
# ---

bind TwoMdiButton <Enter> {
    %W configure -relief raised
}

bind TwoMdiButton <Leave> {
    %W configure -relief flat
}

# ---
# Bindings for frame of slave widget
# ---

# Focus in: make active window
bind TwoMdiBorder <FocusIn> {
    ::two::mdi::FocusIn %W
}

# Motion: set shape of mouse cursor
bind TwoMdiBorder <Motion> {
    ::two::mdi::BorderMotion %W %x %y
}
# Enter: set shape of mouse cursor
bind TwoMdiBorder <Enter> {
    ::two::mdi::BorderMotion %W %x %y
}
# Enter from inside
bind TwoMdiSubFrame <Leave> {
    set b [winfo parent %W] 
    ::two::mdi::BorderMotion $b \
        [expr {%X - [winfo rootx $b]}] \
	[expr {%Y - [winfo rooty $b]}]
}

# Button 1: record cursor and mdi master widget
bind TwoMdiBorder <1> {
    ::two::mdi::BorderB1 %W %x %y
}

# B1-Release: drag finished
bind TwoMdiBorder <B1-ButtonRelease> {
    set ::two::mdi::Mdi(Drag) 0
}

# B1-Motion: resize slave
bind TwoMdiBorder <B1-Motion> {
    ::two::mdi::BorderB1Motion %W %x %X %y %Y
}

# Alternative Impl. to reduce number of redraws
# (Idea taken from a news posting in comp.lang.tcl by Frdric BONNET)
#variable b1motion {}
#bind TwoMdiBorder <B1-Motion> [namespace code {
#    after cancel $b1motion
#    set b1motion [after 50 [namespace code {BorderB1Motion %W %x %X %y %Y}]]
#}]

# right mouse button gives icon menu
bind TwoMdiBorder <3> {
    ::two::mdi::SlavePopup %W %X %Y
}

# ---
# Bindings for title of slave widget
# ---

# Double-One: restore icons and maximize normal slaves
bind TwoMdiTitle <Double-1> {
    ::two::mdi::TitleDouble1 %W
}

# Button 1: retrieve paths of elements of this slave
# and record start of move and mdi master widget
bind TwoMdiTitle <1> {
    ::two::mdi::TitleB1 %W %x %y
}

# B1-Motion: move slave
bind TwoMdiTitle <B1-Motion> {
    ::two::mdi::TitleB1Motion %W %X %Y
}

# B1-Release: move finished; call resize to set scrollbars
# (min and max values may have changed through a widget reaching
# beyond the border)
bind TwoMdiTitle <B1-ButtonRelease> {
    %W configure -cursor left_ptr
    set ::two::mdi::Mdi(Drag) 0
    ::two::mdi::Resize [winfo parent [winfo parent [winfo parent [winfo parent %W]]]]
}

# right mouse button gives icon menu
bind TwoMdiTitle <3> {
    ::two::mdi::SlavePopup [winfo parent [winfo parent [winfo parent %W]]] %X %Y
}

# ---
# Bindings for icon in title of slave widget
# ---

bind TwoMdiIcon <1> {
    ::two::mdi::TitleIcon1 %W %t
}

# ---
# Bindings for container for users's widget
# ---

# if user directly destroys his slave widget:
# clean up things nevertheless
bind TwoMdiSlave <Destroy> {
    ::two::mdi::DeleteSlave [winfo parent [winfo parent %W]] 1
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::mdi::FocusIn --
#	react on focus entering the slave
#
# Arguments:
#	b		Path to slave widget's border

proc ::two::mdi::FocusIn {b} {
    variable Mdi
    # don't do anything if this widget is minimized
    if {!$Mdi($b,min)} {
	set m [winfo parent $b]
	# did we already have an active slave?
	set active $Mdi($m,active)
	# make sure this widget isn't already the active one
	if {![string match $b $active]} {
	    set Mdi($m,active) $b
	    if {[winfo exists $active]} {
        	InactiveSlave $active $m
	    }
	    event generate $m <<ActiveOn>>
	    ActiveSlave $b $m
	    raise $b
	    # make sure that the focus is set to a child
	    # which wants it!
	    SetFocus $b.w.s
	}
    }
}

# ::two::mdi::BorderMotion --
#
#	handle mouse cursor motion events within border frame. Set
#	the mouse cursor shape according to the relative position.
#
# Arguments:
#	b		Path to slave widget's border
#	x		relative X Position of mouse cursor within border (%x)
#	y		relative Y Position (%y)

proc ::two::mdi::BorderMotion {b x y} {
    variable Mdi

    set m [winfo parent $b]
    if {!$Mdi(Drag)} {
        if {$Mdi($m,max) || $Mdi($b,min)} {
       	    $b configure -cursor ""
	} elseif {$x < 10} {
            if {$y < 10} {
        	$b configure -cursor top_left_corner
            } elseif {$y > [winfo height $b] - 10} {
        	$b configure -cursor bottom_left_corner
	    } else {
        	$b configure -cursor left_side
	    }
	} elseif {$x > [winfo width $b] - 10} {
            if {$y < 10} {
        	$b configure -cursor top_right_corner
            } elseif {$y > [winfo height $b] - 10} {
        	$b configure -cursor bottom_right_corner
	    } else {
        	$b configure -cursor right_side
	    }
	} elseif {$y < 10} {
            $b configure -cursor top_side
	} else {
            $b configure -cursor bottom_side
	}
    }
    return
}

# ::two::mdi::BorderB1 --
#
#	handle button press within within border frame.
#
# Arguments:
#	b		Path to slave widget's border
#	x		relative X Position of mouse cursor within border (%x)
#	y		relative Y Position (%y)

proc ::two::mdi::BorderB1 {b x y} {
    variable Mdi

    set m [winfo parent $b]
    if {! ($Mdi($m,max) || $Mdi($b,min))} {
	set Mdi(Drag)     1
	set Mdi(Cursor)   [$b cget -cursor]
	set Mdi(Master,w) [winfo width  $m]
	set Mdi(Master,h) [winfo height $m]
	set Mdi(x)        [expr {[winfo width  $b] - $x}]
	set Mdi(y)        [expr {[winfo height $b] - $y}]
	set Mdi(xx)       [expr {$x + [winfo rootx $m] + [$m cget -bd]}]
	set Mdi(yy)       [expr {$y + [winfo rooty $m] + [$m cget -bd]}]
	set Mdi(Master)   "_two$m"

	SetFocus $b.w.s
    }
    return
}

# ::two::mdi::BorderB1Motion --
#
#	handle interactive resizing of slave. This relies on the mouse
#	cursor's shape set in motion procedure!
#
# Arguments:
#	b		Path to slave widget's border
#	relX		relative X Position of mouse cursor within border (%x)
#	absX		absolute X Positino (%X)
#	relY		relative Y Position (%y)
#	absY		absolute Y Position (%Y)

proc ::two::mdi::BorderB1Motion {b relX absX relY absY} {
    variable Mdi

    set m [winfo parent $b]
    if {$Mdi(Drag) && (! ($Mdi($m,max) || $Mdi($b,min)))} {
	if {[string match "*right*" $Mdi(Cursor)]} {
	    set x [expr {$relX + $Mdi(x)}]
	    if {$x > 20} {
        	set x [expr {($x + [winfo x $b]) < $Mdi(Master,w) ? $x : $Mdi(Master,w) - [winfo x $b]}]
		place $b -width $x
	    }
	} elseif {[string match "*left*" $Mdi(Cursor)]} {
            set x [expr {($absX - $Mdi(xx)) < 0 ? 0 : ($absX - $Mdi(xx))}]
            set w [expr {[winfo width $b] + [winfo x $b] - [$m cget -bd] - $x}]
	    if {$w > 20} {
		place $b -x $x -width $w
		set Mdi($b,x) $x
	    }
	}
	# no elseif: top-right etc.
	if {[string match "*bottom*" $Mdi(Cursor)]} {
	    set y [expr {$relY + $Mdi(y)}]
	    if {$y > 20} {
        	set y [expr {($y + [winfo y $b]) < $Mdi(Master,h) ? $y : $Mdi(Master,h) - [winfo y $b]}]
		place $b -height $y
	    }
	} elseif {[string match "*top*" $Mdi(Cursor)]} {
            set y [expr {($absY - $Mdi(yy)) < 0 ? 0 : ($absY - $Mdi(yy))}]
            set h [expr {[winfo height $b] + [winfo y $b] - [$m cget -bd] - $y}]
	    if {$h > 20} {
		place $b -y $y -height $h
		set Mdi($b,y) $y
	    }
	}
    }
    return
}

# ::two::mdi::TitleB1 --
#
#	handle button press within within title frame.
#
# Arguments:
#	t		Path to slave widget's border
#	x		relative X Position of mouse cursor within border (%x)
#	y		relative Y Position (%y)

proc ::two::mdi::TitleB1 {t x y} {
    variable Mdi

    set b [winfo parent [winfo parent [winfo parent $t]]]
    set m [winfo parent $b]
    if {!$Mdi($m,max) && !$Mdi($b,min)} {
        set Mdi(Drag)      1
	set Mdi(Master)    "_two$m"
	set Mdi(Border)    $b
	set Mdi(Master,x1) [expr {[winfo rootx $m] + [$b cget -bd] - [$m cget -bd]}]
	set Mdi(Master,y1) [expr {[winfo rooty $m] + [$b cget -bd] - [$m cget -bd]}]
	set Mdi(Master,x2) [expr {$Mdi(Master,x1) + [winfo width  $m]}]
	set Mdi(Master,y2) [expr {$Mdi(Master,y1) + [winfo height $m]}]
	set Mdi(xx)        [expr {$x + 2 + [winfo x $t] + $Mdi(borderW) + $Mdi(Master,x1)}]
	set Mdi(yy)        [expr {$y + 2 + [winfo y $t] + $Mdi(borderW) + $Mdi(Master,y1)}]
	$t configure -cursor fleur
    }

    SetFocus $b.w.s
    return
}

# ::two::mdi::TitleB1Motion --
#
#	handle interactive resizing of slave. This relies on the mouse
#	cursor's shape set in motion procedure!
#
# Arguments:
#	t		Path to slave widget's border
#	absX		absolute X Positino (%X)
#	absY		absolute Y Position (%Y)

proc ::two::mdi::TitleB1Motion {t absX absY} {
    variable Mdi

    if {$Mdi(Drag)} {
	# check bounds: cursor within master
	if {$absX < $Mdi(Master,x1)} {
	    set x $Mdi(Master,x1)
	} elseif {$absX > $Mdi(Master,x2)} {
	    set x $Mdi(Master,x2)
	} else {
	    set x $absX
	}

	if {$absY < $Mdi(Master,y1)} {
	    set y $Mdi(Master,y1)
	} elseif {$absY > $Mdi(Master,y2)} {
	    set y $Mdi(Master,y2)
	} else {
	    set y $absY
	}

	place $Mdi(Border) -x [expr {$x - $Mdi(xx)}] -y [expr {$y - $Mdi(yy)}]
    }
    return
}

# ::two::mdi::TitleDouble1 --
#
#	handle doubleclick in title bar of slave.
#
# Arguments:
#	w		Path to Mdi child's title bar

proc ::two::mdi::TitleDouble1 {w} {
    variable Mdi

    set f [winfo parent [winfo parent [winfo parent $w]]]
    set m [winfo parent $f]
    if {$Mdi($f,min)} {
        RestoreMin $f
    } elseif {$Mdi($m,max)} {
        RestoreMax $f
    } else {
        Maximize $f
    }
}

# ::two::mdi::TitleIcon1 --
#
#	handle click on icon in title bar (doubleclick as well)
#
# Arguments:
#	w		Path to icon in Mdi child's title bar
#	t		Time field of click event

proc ::two::mdi::TitleIcon1 {w t} {
    variable Mdi
    set l [winfo parent $w]
    set b [winfo parent [winfo parent $l]]

    if {($Mdi($b,lastmenu) != 0) && ($t - $Mdi($b,lastmenu) < $Mdi(doubleclick))} {
        set Mdi($b,lastmenu) $t
        DeleteSlave $b
    } else {
        set Mdi($b,lastmenu) $t
        SlavePopup $b [winfo rootx $l] [expr {[winfo rooty $l] + [winfo height $l]}]
    }
}

# ::two::mdi::Resize --
#
#	update scrollbars on resize of master
#
# Arguments:
#	m		Path to Mdi master widget

proc ::two::mdi::Resize {m} {
    variable Mdi

    # return if no scrolling is needed
    if {$Mdi($m,max)} {
        return
    }

    # move icons
    if {$Mdi($m,arrangeicons) && [llength $Mdi($m,slaves)]} {
	set Mdi($m,icons) ""
	set iconX 0
	set iconY 0
	set iconH [expr {[winfo reqheight [lindex $Mdi($m,slaves) 0].w.t] \
	                + 2 * ($Mdi(borderW) + $Mdi($m,slavebd))}]
    }

    # claculate bounding box of all slaves and master
    # and position of icons
    set minX 0
    set minY 0
    set maxX [winfo width  $m]
    set maxY [winfo height $m]
    foreach b [AllSlaveBorders $m] {
        if {$Mdi($b,min)} {
	    if {$Mdi($m,arrangeicons)} {
		place $b -x [expr {150 * $iconX}] -y [expr { - $iconH * $iconY}] \
	    	    -rely 1 -width 150 -height $iconH -relwidth 0 -relheight 0 -anchor sw
		incr iconX
		if [expr {150 * ($iconX + 1)  > $maxX}] {
		    # position out of window
		    set  iconX 0
		    incr iconY
		}
	    }
	} elseif {[string length $Mdi($m,xscrollcommand)] || [string length $Mdi($m,yscrollcommand)]} {
	    set x [winfo x $b]
	    set y [winfo y $b]
	    set w [winfo width  $b]
	    set h [winfo height $b]
            if {$x < $minX} {set minX $x}
            if {$y < $minY} {set minY $y}
            if {$x + $w > $maxX} {set maxX [expr {$x + $w}]}
            if {$y + $h > $maxY} {set maxY [expr {$y + $h}]}
	    set Mdi($b,x) $x
	    set Mdi($b,y) $y
	}
    }
    # set value of scrollbars
    set w [winfo width  $m]
    set h [winfo height $m]
    # configure x scroll
    if {[string length $Mdi($m,xscrollcommand)]} {
	set Mdi($m,xOffset)  [expr { $minX < 0  ? 0 - $minX : 0}]
        set Mdi($m,maxWidth) [expr {($maxX > $w ? $maxX     : $w) + $Mdi($m,xOffset)}]

	set minFract [expr { $Mdi($m,xOffset)       / double($Mdi($m,maxWidth))}]
	set maxFract [expr {($Mdi($m,xOffset) + $w) / double($Mdi($m,maxWidth))}]

        eval $Mdi($m,xscrollcommand) [list $minFract $maxFract]
    }
    # configure y scroll
    if {[string length $Mdi($m,yscrollcommand)]} {
	set Mdi($m,yOffset)   [expr { $minY < 0  ? 0 - $minY : 0}]
        set Mdi($m,maxHeight) [expr {($maxY > $h ? $maxY     : $h) + $Mdi($m,yOffset)}]

	set minFract [expr { $Mdi($m,yOffset)       / double($Mdi($m,maxHeight))}]
	set maxFract [expr {($Mdi($m,yOffset) + $h) / double($Mdi($m,maxHeight))}]

        eval $Mdi($m,yscrollcommand) [list $minFract $maxFract]
    }
    return
}

# ::two::mdi::QueueResize --
#
#	Insert a resize into the event loop. If there is already one,
#	it will be removed.
#
# Arguments:
#	m		Path to Mdi master widget

proc ::two::mdi::QueueResize {m} {
    variable Mdi
    
    if {[info exists Mdi($m,resizeQ)]} {
        after cancel $Mdi($m,resizeQ)
    }
    set Mdi($m,resizeQ) [after idle [namespace code "
        [list Resize $m]
        [list unset Mdi($m,resizeQ)]
    "]]
}

# ::two::mdi::Xview --
#
#	set x view on scroll
#
# Arguments:
#	m		Path to Mdi master widget
#	args		Where to scroll to

proc ::two::mdi::Xview {m args} {
    variable Mdi
    # return if no scrolling needed
    if {$Mdi($m,max)} {
        return
    }
    set w [winfo width $m]
    switch -exact -- [lindex $args 0] {
	moveto {
	    set offset [expr {[lindex $args 1] * $Mdi($m,maxWidth)}]
	}
	scroll {
	    switch -exact -- [lindex $args 2] {
	        units {
		    set offset [expr {$Mdi($m,xOffset) + 10 * [lindex $args 1]}]
		}
		pages {
		    set offset [expr {$Mdi($m,xOffset) + [lindex $args 1] * $w / 2}]
		}
		default {
		    return -code error "wrong # args: should be \"$m xview scroll number units|pages\""
		}
	    }
        }
	default {
	    return -code error "unknown option \"[lindex $args 0]\": must be moveto or scroll"
	}
    }
    # suitable range for offset
    if {$offset < 0} {
	set offset 0
    }
    set maxOffset [expr {$Mdi($m,maxWidth) - $w}]
    if {$offset > $maxOffset} {
        set offset $maxOffset
    }
    # move widgets
    foreach b $Mdi($m,slaves) {
	if {!$Mdi($b,min)} {
	    set x [expr {$Mdi($b,x) - $offset + $Mdi($m,xOffset)}]
	    place $b -x $x
	    set Mdi($b,x) $x
	}
    }

    # reset scrollbar position
    set minFract [expr { $offset       / double($Mdi($m,maxWidth))}]
    set maxFract [expr {($offset + $w) / double($Mdi($m,maxWidth))}]
    eval $Mdi($m,xscrollcommand) [list $minFract $maxFract]
    set Mdi($m,xOffset) $offset
    return
}

# ::two::mdi::Yview --
#
#	set y view on scroll
#
# Arguments:
#	m		Path to Mdi master widget
#	args		Where to scroll to

proc ::two::mdi::Yview {m args} {
    variable Mdi
    # return if no scrolling needed
    if {$Mdi($m,max)} {
        return
    }
    set h [winfo height $m]
    switch -exact -- [lindex $args 0] {
	moveto {
	    set offset [expr {[lindex $args 1] * $Mdi($m,maxHeight)}]
	}
	scroll {
	    switch -exact -- [lindex $args 2] {
	        units {
		    set offset [expr {$Mdi($m,yOffset) + 10 * [lindex $args 1]}]
		}
		pages {
		    set offset [expr {$Mdi($m,yOffset) + [lindex $args 1] * $h / 2}]
		}
		default {
		    return -code error "wrong # args: should be \"$m xview scroll number units|pages\""
		}
	    }
        }
	default {
	    return -code error "unknown option \"[lindex $args 0]\": must be moveto or scroll"
	}
    }
    # suitable range for offset
    if {$offset < 0} {
	set offset 0
    }
    set maxOffset [expr {$Mdi($m,maxHeight) - $h}]
    if {$offset > $maxOffset} {
        set offset $maxOffset
    }
    # move widgets
    foreach b $Mdi($m,slaves) {
	if {!$Mdi($b,min)} {
	    set y [expr {$Mdi($b,y) - $offset + $Mdi($m,yOffset)}]
	    place $b -y $y
	    set Mdi($b,y) $y
	}
    }

    # reset scrollbar position
    set minFract [expr { $offset       / double($Mdi($m,maxHeight))}]
    set maxFract [expr {($offset + $h) / double($Mdi($m,maxHeight))}]
    eval $Mdi($m,yscrollcommand) [list $minFract $maxFract]
    set Mdi($m,yOffset) $offset
    return
}

# ::two::mdi::SlavePopup --
#
#	show popup menu for slave window, also used as icon menu.
#	Make sure there is some some displacement from mouse coordinates to
#	avoid problems with grab when clicking in fast sequence
#
# Arguments:
#	b		Path to slave widget's border
#	x, y		Position of popup

proc ::two::mdi::SlavePopup {b x y} {
    variable Mdi
    
    if {$y <= [winfo pointery $b]} {
        set y [expr {[winfo pointery $b] + 1}]
    }

    set m [winfo parent $b]
    set Mdi(Border) $b
    set Mdi(Master) $m

    $m.menu entryconfigure 0 -state \
        [expr {$Mdi($b,min)||$Mdi($m,max) ? "normal" : "disabled"}]
    $m.menu entryconfigure 1 -state \
        [expr {$Mdi($b,min) ? "disabled" : "normal"}]
    $m.menu entryconfigure 2 -state \
        [expr {$Mdi($m,max) ? "disabled" : "normal"}]
    if { [string match $Mdi($m,active) $b] && ([llength [AllSlaveBorders $m 1]] > 1) } {
        $m.menu entryconfigure 3 -state "normal"
    } else {
        $m.menu entryconfigure 3 -state "disabled"
    }
    tk_popup $m.menu $x $y
    return
}

# ::two::mdi::Minimize --
#
#	minimize slave window
#
# Arguments:
#	b		Path to Slave windows's border

proc ::two::mdi::Minimize {b} {
    variable Mdi

    set m [winfo parent $b]

    if {!$Mdi($m,max)} {
        set Mdi($b,place) [place info $b]
    }

    if {$Mdi($m,arrangeicons)} {
	# get stacking order right
	set topIcon ""
	foreach s [AllSlaveBorders $m] {
	    if {$Mdi($s,min)} {
		set topIcon $s
	    }
	}
	if {[string length $topIcon]} {
	    raise $b $topIcon
	} else {
	    lower $b
	}
    } else {
        # search for first free position
	set x   0
	set y   0
	set max [winfo width $m]
	set h   [expr {[winfo height $b.w.t] + 2 * ($Mdi(borderW) + $Mdi($m,slavebd))}]
	while 1 {
	    if {[lsearch -exact $Mdi($m,icons) "$x,$y"] >= 0} {
        	# position already taken
		incr x
	    } elseif {150 * ($x + 1) > $max} {
		# position out of window
		set  x 0
		incr y
	    } else {
		# found pos
		lappend Mdi($m,icons) "$x,$y"
		set     Mdi($b,icon)  "$x,$y"
		# set x and y to widget coordinates
		set x [expr {150 * $x}]
		set y [expr {- $h * $y}]
		break
	    }
	}
 	place $b -x $x -y $y -rely 1 -width 150 -height $h -relwidth 0 -relheight 0 -anchor sw
 	lower $b
    }

    # flag current widget as minimized
    set Mdi($b,min) 1

    # take away focus
    if {[string match $b $Mdi($m,active)]} {
        # $b is already marked as minimized and won't show up in the following list
        set next [lindex [AllSlaveBorders $m 1] end]
	if {[string length $next]} {
            SetFocus $next.w.s
	    # nothing more to do - the <FocusIn> binding will care about state
	} else {
	    # there ain't no remaining slave
	    # and we don't want a <FocusOut> binding, so handle this case here...
            # Find next widget which is willing to take the focus
	    set next [tk_focusNext $m]
	    # Is this, for any reason, within the mdi? Shouldn't be, as there
	    # is really no slave remaining which should take the focus
	    if {[string match "$m*" $next]} {
	        focus [winfo toplevel $m]
	    } else {
	        focus $next
	    }
	    InactiveSlave $b $m
	    set Mdi($m,active) ""
	    event generate $m <<ActiveOff>>
	}
    }

    $b.w.t.min configure -image "$Mdi($m,icon)Rest" \
	-command [namespace code "RestoreMin $b"]
    $b.w.t.max configure -image "$Mdi($m,icon)Max" \
	-command [namespace code "Maximize $b"]

    if {$Mdi($m,max)} {
        RestoreMax $b
    }
    # recalculate scrollbar settings
    # (in arrangeicons mode also icon positons)
    QueueResize $m
    return
}

# ::two::mdi::Maximize --
#
#	maximize slave window (and all others)
#
# Arguments:
#	b		Path to slave windows's border

proc ::two::mdi::Maximize {b} {
    variable Mdi

    set m [winfo parent $b]

    if {$Mdi($m,max)} {
        # maximize when already maximized can happen through
	# "master show slave"
	# in this case we only have to work on this slave
	set candidates $b
    } else {
        # maximize from normal state:  work on all normal
	# (i.e. non-minimized) slaves
        set candidates $Mdi($m,slaves)
	set Mdi($m,max) 1
    }

    foreach s $candidates {
	if {!$Mdi($s,min)} {
	    set Mdi($s,place) [place info $s]
	}
	if {$Mdi($s,min) && $s == $b} {
	    set Mdi($s,min) 0
	}
	if {!$Mdi($s,min)} {
	    place $s -x 0 -y 0 -rely 0 -width 0 -height 0 -relwidth 1 -relheight 1 -anchor nw
	    $s.w.t.min configure -image "$Mdi($m,icon)Min" \
		-command [namespace code "Minimize $s"]
	    $s.w.t.max configure -image "$Mdi($m,icon)Rest" \
		-command [namespace code "RestoreMax $s"]
        }
    }
    
    # give focus
    SetFocus $b.w.s

    # recalculate scrollbar settings
    if {[string length $Mdi($m,yscrollcommand)]} {
        eval $Mdi($m,yscrollcommand) 0 1
    }
    if {[string length $Mdi($m,xscrollcommand)]} {
        eval $Mdi($m,xscrollcommand) 0 1
    }
    return
}

# ::two::mdi::Restore --
#
#	restore slave window to former size. Calls RestoreMin
#	or RestoreMax, depending on current state.
#
# Arguments:
#	b		Path to slave windows's border

proc ::two::mdi::Restore {b} {
    variable Mdi
    
    if {$Mdi($b,min)} {
	RestoreMin $b
    } elseif {$Mdi([winfo parent $b],max)} {
	RestoreMax $b
    }
    return
}

# ::two::mdi::RestoreMin --
#
#	restore size and position of slave window.
#	Call this when restoring from minimized state.
#
# Arguments:
#	b		Path to slave windows's border

proc ::two::mdi::RestoreMin {b} {
    variable Mdi

    set m [winfo parent $b]

    # remove from list of icons
    if {!$Mdi($m,arrangeicons)} {
        set icon [lsearch -exact $Mdi($m,icons) $Mdi($b,icon)]
        set Mdi($m,icons) [lreplace $Mdi($m,icons) $icon $icon]
    }

    eval [list place configure $b] $Mdi($b,place)
    $b.w.t.min configure -image "$Mdi($m,icon)Min" \
	-command [namespace code "Minimize $b"]
    $b.w.t.max configure -image "$Mdi($m,icon)Max" \
	-command [namespace code "Maximize $b"]
    set Mdi($b,min) 0

    # give focus; first, set focus to toplevel in order to
    # really trigger FocusIn binding. If focus were already,
    # for whatever reason, inside the minimized slave, things
    # would go boom...
    focus [winfo toplevel $m]
    SetFocus $b.w.s
    
    # Make sure that the mdi master doesn't think we are in maximized
    # mode: this could happen if we had indeed some maximized slaves
    # and the were all closed. Opening new windows should put them in
    # maximized state, so this is o.k., but if we want to restore a
    # minimized window, then forget about maximization.
    set Mdi($m,max) 0
    # reset scrollbar values
    QueueResize $m
    return
}

# ::two::mdi::RestoreMax --
#
#	restore size and place of slave window and maybe all others.
#	Call this when restoring from maximized state.
#
# Arguments:
#	b		Path to slave windows's border

proc ::two::mdi::RestoreMax {b} {
    variable Mdi

    set m [winfo parent $b]
    set Mdi($m,max) 0

    foreach b $Mdi($m,slaves) {
	if {!$Mdi($b,min)} {
            eval [list place configure $b] $Mdi($b,place)
	    $b.w.t.min configure -image "$Mdi($m,icon)Min" \
	        -command [namespace code "Minimize $b"]
	    $b.w.t.max configure -image "$Mdi($m,icon)Max" \
	        -command [namespace code "Maximize $b"]
	}
    }
    # reset scrollbar values
    QueueResize $m
    return
}

# ::two::mdi::LowerSlave --
#
#	move slave to bottom of stacking order.
#	Deals with focus: FocusIn binding may trigger after popupmenu
#	and raise the window.
#
# Arguments:
#	b		Path to slave window's border

proc ::two::mdi::LowerSlave {b} {
    set m  [winfo parent $b]
    set bb [AllSlaveBorders $m 1]
    set l  [llength $bb]

    if {$l > 1} {
	lower $b
	SetFocus [lindex $bb [expr {$l - 2}]].w.s
    }
    return
}

# ::two::mdi::Shuffle --
#
#	modify stacking order. Shuffle up will turn lowest slave to top,
#	while Shuffle down will lower the toplevel
#
# Arguments:
#	dir		direction: "up" or "down"
#	m		Path to master widget

proc ::two::mdi::Shuffle {dir m} {
    variable Mdi

    set bb [AllSlaveBorders $m 1]
    set l  [llength $bb]
    if {$l > 1} {
	if {[string match "up" $dir]} {
            # don't use $Mdi($m,slaves): here we take advantage of
	    # of the stacking order of the slaves!
	    SetFocus "[lindex $bb 0].w.s"
	} elseif {[string match "down" $dir]} {
	    LowerSlave [lindex $bb [expr {$l - 1}]]
	} else {
            return -code error "unknown option \"$dir\": must be up or down"
	}
    }
    return
}

# ::two::mdi::NoScrolling --
#
#	Resize all slaves and update geometry in order to disable
#	Scrolling in Mdi master. This is needed for commands that
#	know that they stack all slaves in a way that will remove
#	the need for scrolling and needs to know the size of the
#	Master *after* that change. Scrollbars might be automatic,
#	so we really have to make the change and see how large the
#	Mdi is afterwards.
#
# Arguments:
#	m		master widget
#	bb		List of all affected slaves

proc ::two::mdi::NoScrolling {m bb} {
    variable Mdi

    # Any scrollbars around? And if so, is there a changes for them
    # to really scroll the Mdi?
    if {([string length $Mdi($m,xscrollcommand)] &&
	   ($Mdi($m,yOffset) != 0 ||
	    $Mdi($m,maxWidth) > [winfo width $m])
    )|| ([string length $Mdi($m,yscrollcommand)] &&
	   ($Mdi($m,xOffset) != 0 ||
	    $Mdi($m,maxHeight) > [winfo height $m])
    )} {
	# O.k., we'll have to take make sure that all geometry is updated
	# as if the slaves were smaller than the Master. This is needed
	# in case we have auto scrollbars which would vanish *after*
	# calculating the size the slave and thus leave a gap behind.

        # Settings for all slaves:
    	foreach b $bb {
	    set Mdi($b,x) 0
	    set Mdi($b,y) 0
	    place $b -x 0 -y 0 -width 1 -height 1
	}
	# recalculate scrollbar settings
	if {[string length $Mdi($m,yscrollcommand)]} {
	    set Mdi($m,yOffset)   0
            set Mdi($m,maxHeight) [winfo height $m]
            eval $Mdi($m,yscrollcommand) 0 1
	}
	if {[string length $Mdi($m,xscrollcommand)]} {
	    set Mdi($m,xOffset)  0
            set Mdi($m,maxWidth) [winfo width $m]
            eval $Mdi($m,xscrollcommand) 0 1
	}
	update idletasks
    }
    return
}

# ::two::mdi::Tile --
#
#	Tile all slave windows in master.
#	1-3 slaves: place one above the other
#	n*3 slaves: place in n columns
#	other: fill last column starting from the bottom
#
# Arguments:
#	m		master widget

proc ::two::mdi::Tile {m} {
    variable Mdi

    # get list of all slaves; don't use $Mdi($m,slaves)
    # because we use the stacking order of the slaves!
    set bb     [AllSlaveBorders $m 1]
    set count  [llength $bb]
    if {$count > 0} {
	if {$Mdi($m,max)} {
	    RestoreMax [lindex $bb 0]
	}

	NoScrolling $m $bb

	set width  [expr {[winfo width  $m] - 1}]
	set height [expr {[winfo height $m] - 1}]
	set rows   [expr {$count >= 3 ? 3 : $count}]
	set cols   [expr {int( ($count + 2) / 3 ) }]
	set add    [expr {$rows * $cols - $count}]
	set h      [expr {$height / $rows}]
        set w      [expr {$width / ($add > 0 ? $cols - 1 : $cols)}]

	set x 0
	set y 0
	for {set i [expr {$count - 1}]} {$i >= 0} {incr i -1} {
	    set b [lindex $bb $i]
	    if {$x == 0 && $y == $add} {
		set w [expr {$width / $cols} ]
		incr cols
	    }
	    set Mdi($b,x) [expr {$x * $w}]
	    set Mdi($b,y) [expr {$y * $h}]
	    place $b -x $Mdi($b,x) -y $Mdi($b,y) -width $w -height $h
	    incr x
	    if {$x >= $cols - 1} {
		set x 0
		incr y
	    }
	}
    }
    return
}

# ::two::mdi::Cascade --
#
#	Cascade the slave widgets (stack them with some displacement).
#
# Arguments:
#	m		master widget

proc ::two::mdi::Cascade {m} {
    variable Mdi

    # get list of all slaves; don't use $Mdi($m,slaves):
    # we need the stacking order of the slaves
    set bb     [AllSlaveBorders $m 1]
    set count  [llength $bb]

    if {$count > 0} {
	if {$Mdi($m,max)} {
	    RestoreMax [lindex $bb 0]
	}

	NoScrolling $m $bb

	# calculate width and height of slave
	# each should be 75% of master, but at least have a minimum size
	set width  [winfo width  $m]
	set height [winfo height $m]
	set slaveW [expr {0.75 * $width}]
	set slaveH [expr {0.75 * $height}]
	if {$slaveW < $Mdi(slaveW)} {
            set slaveW $Mdi(slaveW)
	}
	if {$slaveH < $Mdi(slaveH)} {
            set slaveH $Mdi(slaveH)
	}

        set offset $Mdi(offset0)
	foreach b $bb {
            place $b -x $offset -y $offset -width $slaveW -height $slaveH
	    incr offset $Mdi(step)
            if {$offset > $width - $slaveW || $offset > $height - $slaveH } {
        	set offset $Mdi(offset0)
	    }
	}
    }
    QueueResize $m
    return
}

# ::two::mdi::AllSlaveBorders --
#
#	Returns list of all slaves of master widget. If $noMin is set to true,
#	then only normal slaves (no matter if maximized or not) are included.
#
# Arguments:
#	m		master
#	noMin		don't include minimized slaves if set to true

proc ::two::mdi::AllSlaveBorders {m {noMin 0}} {
    variable Mdi

    set bb ""
    foreach b [winfo children $m] {
        if {![string match $b "$m.menu"] &&
	    ![string match $b "$m.root"] &&
	    !($noMin && $Mdi($b,min))
	} {
	    lappend bb $b
	}
    }
    return $bb
}


# ::two::mdi::ConfigureSlave --
#
#	Configure settings according to Mdi($m,...) for slave widget
#
# Arguments:
#	b		Path to Slave window's border

proc ::two::mdi::ConfigureSlave {b} {
    variable Mdi

    # Master
    set m [winfo parent $b]

    # Frame
    $b configure -bg $Mdi($m,slavebg) -bd $Mdi($m,slavebd)

    # Title: active slave?
    if {[string match $b $Mdi($m,active)]} {
        ActiveSlave $b $m
    } else {
        InactiveSlave $b $m
    }
    # Title: font and text
    if {[string length $Mdi($m,font)]} {
	$b.w.t.label configure -text $Mdi($b,title) -font $Mdi($m,font)
    } else {
	$b.w.t.label configure -text $Mdi($b,title)
    }

    # Spaceing
    $b.w.d configure -bg $Mdi($m,slavebg)
    return
}

# ::two::mdi::SetSlavePos --
#
#	Set the position of a slave widget according to the
#	spec as generated by entrycget -pos. This spec is a
#	list: first element is the state {min|max|normal}, the
#	second is the output of [place info] for the border
#	frame (for the normal state).
#
# Arguments:
#	m		Path to mdi master widget
#	b		Path to Slave window's border
#	spec		Spec for state and position

proc ::two::mdi::SetSlavePos {m b spec} {
    variable Mdi

    set state [lindex $spec 0]
    set pos   [lindex $spec 1]
    if {[string match "min" $state] && !$Mdi($b,min)} {
        Minimize $b
    } elseif {[string match "max" $state] && !$Mdi($m,max)} {
        Maximize $b
    }
    if {[string match "min" $state] || [string match "max" $state] } {
        set Mdi($b,place) $pos
    } else {
        eval [list place configure $b] $pos
    }
    array set conf $pos
    set Mdi($b,x) $conf(-x)
    set Mdi($b,y) $conf(-y)
    QueueResize $m
}


# ::two::mdi::SetSlaveStyle --
#
#	sets slave's style according to the style (windows/unix)
#
# Arguments:
#	b		Path to Slave window's border

proc ::two::mdi::SetSlaveStyle {b} {
    variable Mdi

    set m  [winfo parent   $b]
    set tl [winfo toplevel $m]
    
    if {[string match "unix" $Mdi($m,style)]} {
        bindtags $b.w.t.close [list $b.w.t.close TwoMdiButton Button $tl all]
        bindtags $b.w.t.max   [list $b.w.t.max   TwoMdiButton Button $tl all]
        bindtags $b.w.t.min   [list $b.w.t.min   TwoMdiButton Button $tl all]
	$b.w.t.close configure -relief flat -image mdiKdeClose
	$b.w.t.max   configure -relief flat
	$b.w.t.min   configure -relief flat
        pack $b.w.t.close -side right -padx 0 -pady 0 -fill y
        pack $b.w.t.max   -side right         -pady 0 -fill y
        pack $b.w.t.min   -side right         -pady 0 -fill y
        pack $b.w.t.icon  -side left  -padx 2 -pady 0 -fill y
        pack $b.w.t.label -side left -expand 1              -fill both
    } else {
        bindtags $b.w.t.close [list $b.w.t.close Button $tl all]
        bindtags $b.w.t.max   [list $b.w.t.max   Button $tl all]
        bindtags $b.w.t.min   [list $b.w.t.min   Button $tl all]
	$b.w.t.close configure -relief raised -image mdiWinClose
	$b.w.t.max   configure -relief raised
	$b.w.t.min   configure -relief raised
        pack $b.w.t.close -side right -padx 2 -pady 2 -fill none
        pack $b.w.t.max   -side right         -pady 2 -fill none
        pack $b.w.t.min   -side right         -pady 2 -fill none
        pack $b.w.t.icon  -side left  -padx 2 -pady 2 -fill y
        pack $b.w.t.label -side left -expand 1              -fill both
    }

    if {$Mdi($m,max)} {
	$b.w.t.min configure -image "$Mdi($m,icon)Min"
	$b.w.t.max configure -image "$Mdi($m,icon)Rest"
    } elseif {$Mdi($b,min)} {
	$b.w.t.min configure -image "$Mdi($m,icon)Rest"
	$b.w.t.max configure -image "$Mdi($m,icon)Max"
    } else {
	$b.w.t.max configure -image "$Mdi($m,icon)Max"
	$b.w.t.min configure -image "$Mdi($m,icon)Min"
    }
    return
}

# ::two::mdi::ActiveSlave --
#
#	Configure color settings to show active state of slave
#
# Arguments:
#	b		Path to slave window's border
#	m		Path to master

proc ::two::mdi::ActiveSlave {b m} {
    variable Mdi
    $b.w.t       configure -bg $Mdi($m,activebg)
    $b.w.t.icon  configure -bg $Mdi($m,activebg)
    $b.w.t.label configure -bg $Mdi($m,activebg) -fg $Mdi($m,activefg)
    return
}

# ::two::mdi::InactiveSlave --
#
#	Configure color settings to show inactive state of slave
#
# Arguments:
#	b		Path to slave window's border
#	m		Path to master

proc ::two::mdi::InactiveSlave {b m} {
    variable Mdi
    $b.w.t       configure -bg $Mdi($m,titlebg)
    $b.w.t.icon  configure -bg $Mdi($m,titlebg)
    $b.w.t.label configure -bg $Mdi($m,titlebg) -fg $Mdi($m,titlefg)
    return
}

# ::two::mdi::SetFocus --
#
#	Set focus to the first child of $w which takes focus
#	or, if there is none, to $w. If any child of $w already
#	has the focus: don't mess with it.
#
# Arguments:
#	w		Path to widget

proc ::two::mdi::SetFocus {w} {
    if {![string match "$w.*" [focus]]} {
	set child [FocusNextChild $w]
	if {[string match "$w*" $child]} {
            focus $child
	} else {
	    focus $w
	}
    }
    return
}

# ::two::mdi::FocusNextChild --
#	Look for the first child of the widget "w" which is willing
#	to take the focus (like in keyboard traversal). Skip
#	toplevels!
#	(modified version of tk_focusNext from tk 8.0.3)
#
# Arguments:
#	w		Path to widget

proc ::two::mdi::FocusNextChild {w} {
    set cur $w
    while {1} {

	# Descend to just before the first child of the current widget.

	set parent $cur
	set children [winfo children $cur]
	set i -1

	# Look for the next sibling that isn't a top-level.

	while {1} {
	    incr i
	    if {$i < [llength $children]} {
		set cur [lindex $children $i]
		if {[string compare [winfo toplevel $cur] $cur] == 0} {
		    continue
		} else {
		    break
		}
	    }

	    # No more siblings, so go to the current widget's parent.
	    # If it's the current widget, break out of the loop, otherwise
	    # look for its next sibling.

	    set cur $parent
	    if {[string compare $w $cur] == 0} {
		break
	    }
	    set parent [winfo parent $parent]
	    set children [winfo children $parent]
	    set i [lsearch -exact $children $cur]
	}
	if {[string compare $cur $w] == 0} {
	    return $cur
	}
	set errCode [catch {$cur cget -takefocus} focusOk]
	if {![string length $focusOk]} {
	    set focusOk 1
	}
	# This is a bit of a HACK: a TwoEditor widget might be
	# split, and then one of the clones is active, i.e. had the
	# focus before the whole editor lost the focus.
	if {[string match TwoEditor [$cur cget -class]]} {
	    return $::two::editor::Editor($cur,active)
	}
	if {!$errCode && $focusOk} {
	    return $cur
	}
    }
}

# ::two::mdi::Animate --
#
#	show anomation for window size changes (minimize/maximize/restore)
#	BETTER NOT USE IT - it's too slow :-(
#
# Arguments:
#	m		Master widget
#	x1		start x position
#	y1		start y position
#	w1		start width
#	h1		start height
#	x2,y2,w2,h2	dito for end

proc ::two::mdi::Animate {m x1 y1 w1 h1 x2 y2 w2 h2} {
    variable Mdi
    set a $m._anim
    if {[winfo exists $a]} {
        return
    }
    frame $a -bg $Mdi($m,activebg)
    set dx [expr {($x2 - $x1) / 5}]
    set dy [expr {($y2 - $y1) / 5}]
    set dw [expr {($w2 - $w1) / 5}]
    set dh [expr {($h2 - $h1) / 5}]

    after idle [namespace code _Animate] $a $x1 $y1 $dx $dy $w1 $h1 $dw $dh 4
    return
}

# ::two::mdi::_Animate --
#
#	helper function for aninate.

proc ::two::mdi::_Animate {a x1 y1 dx dy w1 h1 dw dh count} {
    if {$count == 0} {
        destroy $a
    } else {
	incr x1 $dx
	incr y1 $dy
	incr w1 $dw
	incr h1 $dh
	incr count -1
	place $a -x $x1 -y $y1 -width $w1 -height $h1
	update
	after idle [namespace code _Animate] $a $x1 $y1 $dx $dy $w1 $h1 $dw $dh $count
    }
    return
}

# ::two::mdi::DeleteSlave --
#
#	Delete slave widget from MDI master. Clean up $Mdi.
#	Maybe ask closecmd beforehand.
#	Returns 1 if really deleted
#
# Arguments:
#	b		Path to slave window's border
#	force		If set to 1, no closecmd will be called

proc ::two::mdi::DeleteSlave {b {force 0}} {
    variable Mdi

    set m [winfo parent $b]

    # run closecmd if there is one
    if {!$force && [string length $Mdi($b,closecmd)]} {
        set retval [namespace eval :: $Mdi($b,closecmd) [list $b.w.s]]
        if {![regexp -nocase {^(1|true|yes|ok)$} $retval]} {
	    return 0
	}
    }

    # remove from list of slaves
    set pos [lsearch -exact $Mdi($m,slaves) $b]
    set Mdi($m,slaves) [lreplace $Mdi($m,slaves) $pos $pos]

    # Disable the first two root menu entries, i.e. the shuffle operations
    # stop making sense?
    if {[llength $Mdi($m,slaves)] == 1} {
        $m.root entryconfigure 0 -state disabled
        $m.root entryconfigure 1 -state disabled
    }

    if {!$Mdi($m,arrangeicons) && $Mdi($b,min)} {
	# remove from list of icons
	set icon [lsearch -exact $Mdi($m,icons) $Mdi($b,icon)]
	set Mdi($m,icons) [lreplace $Mdi($m,icons) $icon $icon]
    }

    # send an event
    event generate $b.w.s <<Closeing>>

    # remove binding from inner container widget to avoid infinite loop
    # (would call DeleteSlave:-)
    bindtags $b.w.s nothing

    # finally destroy widget
    place forget $b
    destroy $b
    
    # set focus if needed
    if {[string match "." [focus]]} {
        set next [lindex [AllSlaveBorders $m 1] end]
	if {[string length $next]} {
	    # yes, there is another slave to recieve the focus
            SetFocus $next.w.s
	} else {
	    # sorry, this was the last slave which is willing to take focus
	    set Mdi($m,active) ""
            event generate $m <<ActiveOff>>
	}
    } else {
        # No need to reset focus, but this might as well
        # have been the last slave willing to take the focus:-)
        if {[llength [AllSlaveBorders $m 1]] < 1} {
            event generate $m <<ActiveOff>>
        }
    }
    # maybe reorder icons
    if {$Mdi($m,arrangeicons) && $Mdi($b,min)} {
        Resize $m
    }
    # clear Mdi
    foreach key [array names Mdi "$b,*"] {
        unset Mdi($key)
    }

    # resize: now we might have to set the scrollbars different
    Resize $m

    return 1
}

# ::two::mdi::AddSlave --
#
#	Add slave widget to MDI master
#	Widget Hierarchy:
#	$m                          Frame:  master widget
#	+- $m.#                     Frame:  border widget (for internal use); # = nonce = $b
#	|  +- $m.#.w                Frame:  provide space (padding!) for border
#	|  |  +- $m.#.w.t           Frame:  title bar
#	|  |  |  +- $m.#.w.t.icon   Label:  icon in title bar
#	|  |  |  +- $m.#.w.t.label  Label:  text in title bar
#	|  |  |  +- $m.#.w.t.min    Button: minimize or restore from min
#	|  |  |  +- $m.#.w.t.max    Button: maximize or restore from max
#	|  |  |  +- $m.#.w.t.close  Button: close slave
#	|  |  +- $m.#.w.d           Frame:  spaceing (distance) between title and content
#	|  |  +- $m.#.w.s           Frame:  container for real slave = $s (returned to user)
#	+- $m.menu                  Menu:   Slave window menu
#	+- $m.root                  Menu:   Master root menu
#
# Arguments:
#	m		Path to master
#	args		Additional args

proc ::two::mdi::AddSlave {m args} {
    variable Mdi

    # call resize to reset scrollbar values
    # Otherwise we would have to take the current displacement into account
    Resize $m

    # Name of new border
    set b $m.$Mdi(nonce)
    lappend Mdi($m,slaves) $b
    
    # Enable the first two root menu entries, i.e. the shuffle operations
    # start making sense?
    if {[llength $Mdi($m,slaves)] == 2} {
        $m.root entryconfigure 0 -state normal
        $m.root entryconfigure 1 -state normal
    }

    # Set defaults
    set Mdi($b,title)    "New $Mdi(nonce)"
    set Mdi($b,closecmd) ""
    set Mdi($b,min)      0
    set Mdi($b,lastmenu) 0

    # See if there are some args for us
    two_parseArgs args Mdi($b,title)    "-title"
    two_parseArgs args Mdi($b,closecmd) "-closecmd"
    two_parseArgs args posSpec		"-pos"

    # window border is built by a simple frame
    frame $b -class TwoMdi
    # with our additinal bindings
    set tl [winfo toplevel $m]
    bindtags $b [list $b TwoMdiBorder TwoMdiFrame $tl all]

    # another frame within window border to glue components
    # and provide padding space (to show the border)
    pack [frame $b.w -bd 0 -class TwoMdi] \
        -side bottom -expand true -fill both \
        -padx $Mdi(borderW) -pady $Mdi(borderW)
    bindtags $b.w [list $b.w TwoMdiSubFrame $tl all]

    # title
    pack [frame  $b.w.t -bd 0 -class TwoMdiTitle] -side top -fill x
    button $b.w.t.close -takefocus 0 -command [namespace code "DeleteSlave $b"]
    button $b.w.t.max   -takefocus 0 -command [namespace code "Maximize $b"]
    button $b.w.t.min   -takefocus 0 -command [namespace code "Minimize $b"]
    label  $b.w.t.icon  -image mdiText
    label  $b.w.t.label -justify left -anchor w -bd 0
    # frame gets no bindings (Problem: class name equals binding)
    bindtags $b.w.t nothing
    # label and icon get our own bindings
    bindtags $b.w.t.label [list $b.w.t.label TwoMdiTitle $tl all]
    bindtags $b.w.t.icon  [list $b.w.t.icon  TwoMdiIcon  $tl all]
    # configure and pack the elements according to the current style
    SetSlaveStyle $b
    
    # spaceing
    pack [frame $b.w.d -height $Mdi(borderW) -bd 0 -class TwoMdi] -side top -fill x
    bindtags $b.w.d nothing
    
    # slave: container for user's widget
    set s $b.w.s
    pack [uplevel 2 [list frame $s -bd 0 -class TwoMdiSlave] $args] \
        -side bottom -expand true -fill both
    bindtags $s [list $s TwoMdiSlave $tl all]
    
    # configure slave widget
    ConfigureSlave $b

    # do we have a position?
    if {[info exists posSpec]} {
        SetSlavePos $m $b $posSpec
    } else {
        # calculate width and height of slave
        # each should be 75% of master, but at least have a minimum size
        set width  [winfo width  $m]
        set height [winfo height $m]
        set slaveW [expr {0.75 * $width}]
        set slaveH [expr {0.75 * $height}]
        if {$slaveW < $Mdi(slaveW)} {
            set slaveW $Mdi(slaveW)
        }
        if {$slaveH < $Mdi(slaveH)} {
            set slaveH $Mdi(slaveH)
        }
    
        # calculate new offset
        if {$Mdi(offset) > $width - $slaveW || $Mdi(offset) > $height - $slaveH } {
            set Mdi(offset) $Mdi(offset0)
        }
    
        # set new offset as widget position for scrollbar calculation
        set Mdi($b,x) $Mdi(offset)
        set Mdi($b,y) $Mdi(offset)
        
        # place window within master
        if {$Mdi($m,max)} {
            set Mdi($b,place) "-x $Mdi(offset) -y $Mdi(offset) -rely 0 \
                -width $slaveW -height $slaveH -relwidth 0 -relheight 0 -anchor nw"
            place $b -x 0 -y 0 -rely 0 -width 0 -height 0 -relwidth 1 -relheight 1 -anchor nw
            $b.w.t.min configure -image "$Mdi($m,icon)Min" \
                -command [namespace code "Minimize $b"]
            $b.w.t.max configure -image "$Mdi($m,icon)Rest" \
                -command [namespace code "RestoreMax $b"]
        } else {
            place $b -x $Mdi(offset) -y $Mdi(offset) -width $slaveW -height $slaveH
        }
    }
    # resize again: now we might have to set the scrollbars differently
    QueueResize $m

    incr Mdi(offset) $Mdi(step)
    incr Mdi(nonce)
    return $s
}

namespace import ::two::mdi::*
